# -*- coding: utf-8 -*-
"""
xform2json module - Transform an XForm to a JSON dictionary.
"""
from __future__ import print_function

import codecs
import copy
import json
import logging
import re
import xml.etree.ElementTree as ETree
from operator import itemgetter

from pyxform import builder
from pyxform.utils import NSMAP, basestring

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


# {{{ http://code.activestate.com/recipes/573463/ (r7)
class XmlDictObject(dict):
    """
    Adds object like functionality to the standard dictionary.
    """

    def __init__(self, initdict=None):
        if initdict is None:
            initdict = {}
        dict.__init__(self, initdict)

    def __getattr__(self, item):
        return self.__getitem__(item)

    def __setattr__(self, item, value):
        self.__setitem__(item, value)

    def __str__(self):
        if "_text" in self:
            return self.__getitem__("_text")
        else:
            return ""

    @staticmethod
    def wrap(x):
        """
        Static method to wrap a dictionary recursively as an XmlDictObject
        """

        if isinstance(x, dict):
            return XmlDictObject(
                (k, XmlDictObject.Wrap(v)) for (k, v) in iter(x.items())
            )
        elif isinstance(x, list):
            return [XmlDictObject.Wrap(v) for v in x]
        else:
            return x

    @staticmethod
    def _un_wrap(x):
        if isinstance(x, dict):
            return dict((k, XmlDictObject._un_wrap(v)) for (k, v) in iter(x.items()))
        elif isinstance(x, list):
            return [XmlDictObject._un_wrap(v) for v in x]
        else:
            return x

    def un_wrap(self):
        """
        Recursively converts an XmlDictObject to a standard dictionary
        and returns the result.
        """

        return XmlDictObject._un_wrap(self)


def _convert_dict_to_xml_recurse(parent, dictitem):
    assert not isinstance(dictitem, list)

    if isinstance(dictitem, dict):
        for (tag, child) in iter(dictitem.items()):
            if str(tag) == "_text":
                parent.text = str(child)
            elif isinstance(child, list):
                # iterate through the array and convert
                for listchild in child:
                    elem = ETree.Element(tag)
                    parent.append(elem)
                    _convert_dict_to_xml_recurse(elem, listchild)
            else:
                elem = ETree.Element(tag)
                parent.append(elem)
                _convert_dict_to_xml_recurse(elem, child)
    else:
        parent.text = str(dictitem)


def convert_dict_to_xml(xmldict):
    """
    Converts a dictionary to an XML ElementTree Element
    """

    roottag = xmldict.keys()[0]
    root = ETree.Element(roottag)
    _convert_dict_to_xml_recurse(root, xmldict[roottag])
    return root


def _convert_xml_to_dict_recurse(node, dictclass):
    nodedict = dictclass()

    if len(node.items()) > 0:
        # if we have attributes, set them
        nodedict.update(dict(node.items()))

    for child in node:
        # recursively add the element's children
        newitem = _convert_xml_to_dict_recurse(child, dictclass)
        # if tag in between text node, capture the tail end
        if child.tail is not None and child.tail.strip() != "":
            newitem["tail"] = child.tail
        if child.tag in nodedict:
            # found duplicate tag, force a list
            if isinstance(nodedict[child.tag], list):
                # append to existing list
                nodedict[child.tag].append(newitem)
            else:
                # convert to list
                nodedict[child.tag] = [nodedict[child.tag], newitem]
        else:
            # only one, directly set the dictionary
            nodedict[child.tag] = newitem

    if node.text is None:
        text = ""
    else:
        text = node.text.strip()

    if len(nodedict) > 0:
        # if we have a dictionary
        # add the text as a dictionary value (if there is any)
        if len(text) > 0:
            nodedict["_text"] = text
    else:
        # if we don't have child nodes or attributes, just set the text
        nodedict = text

    return nodedict


def convert_xml_to_dict(root, dictclass=XmlDictObject):
    """
    Converts an XML file or ElementTree Element to a dictionary
    """
    # If a string is passed in, try to open it as a file
    if isinstance(root, basestring):
        root = _try_parse(root)
    elif not isinstance(root, ETree.Element):
        raise TypeError("Expected ElementTree.Element or file path string")

    return dictclass({root.tag: _convert_xml_to_dict_recurse(root, dictclass)})


# end of http://code.activestate.com/recipes/573463/ }}}


def _try_parse(root, parser=None):
    """
    Try to parse the root from a string or a file/file-like object.
    """
    root = root.encode("UTF-8")
    try:
        parsed_root = ETree.fromstring(root, parser)
    except ETree.ParseError:
        parsed_root = ETree.parse(root, parser=parser).getroot()
    return parsed_root


class XFormToDict:
    def __init__(self, root):
        if isinstance(root, basestring):
            parser = ETree.XMLParser(encoding="UTF-8")
            self._root = _try_parse(root, parser)
            self._dict = XmlDictObject(
                {
                    self._root.tag: _convert_xml_to_dict_recurse(
                        self._root, XmlDictObject
                    )
                }
            )
        elif not isinstance(root, ETree.Element):
            raise TypeError("Expected ElementTree.Element or file path string")

    def get_dict(self):
        json_str = json.dumps(self._dict)
        for uri in NSMAP.values():
            json_str = json_str.replace("{%s}" % uri, "")
        return json.loads(json_str)


def create_survey_element_from_xml(xml_file):
    sb = XFormToDictBuilder(xml_file)
    return sb.survey()


class XFormToDictBuilder:
    """Experimental XFORM xml to XFORM JSON"""

    QUESTION_TYPES = {
        "select": "select all that apply",
        "select1": "select one",
        "int": "integer",
        "dateTime": "datetime",
        "string": "text",
    }

    def __init__(self, xml_file):
        doc_as_dict = XFormToDict(xml_file).get_dict()
        self._xmldict = doc_as_dict

        assert "html" in doc_as_dict
        assert "body" in doc_as_dict["html"]
        assert "head" in doc_as_dict["html"]
        assert "model" in doc_as_dict["html"]["head"]
        assert "title" in doc_as_dict["html"]["head"]
        assert "bind" in doc_as_dict["html"]["head"]["model"]

        self.body = doc_as_dict["html"]["body"]
        self.model = doc_as_dict["html"]["head"]["model"]
        self.bindings = copy.deepcopy(self.model["bind"])
        self._bind_list = copy.deepcopy(self.model["bind"])
        self.title = doc_as_dict["html"]["head"]["title"]
        self.new_doc = {
            "type": "survey",
            "title": self.title,
            "children": [],
            "id_string": self.title,
            "sms_keyword": self.title,
            "default_language": "default",
        }
        self._set_submission_info()
        self._set_survey_name()
        self.children = []
        self.ordered_binding_refs = []
        self._set_binding_order()

        # set self.translations
        self._set_translations()

        for key, obj in iter(self.body.items()):
            if isinstance(obj, dict):
                self.children.append(self._get_question_from_object(obj, type=key))
            elif isinstance(obj, list):
                for item in obj:
                    self.children.append(self._get_question_from_object(item, type=key))
        self._cleanup_bind_list()
        self._cleanup_children()
        self.new_doc["children"] = self.children

    def _set_binding_order(self):
        self.ordered_binding_refs = []
        for bind in self.bindings:
            self.ordered_binding_refs.append(bind["nodeset"])

    def _set_survey_name(self):
        obj = self.bindings[0]
        name = obj["nodeset"].split("/")[1]
        self.new_doc["name"] = name
        self.new_doc["id_string"] = self.model["instance"][name]["id"]

    def _set_submission_info(self):
        if "submission" in self.model:
            submission = self.model["submission"]
            if "action" in submission:
                self.new_doc["submission_url"] = submission["action"]
            if "base64RsaPublicKey" in submission:
                self.new_doc["public_key"] = submission["base64RsaPublicKey"]
            if "auto-send" in submission:
                self.new_doc["auto_send"] = submission["auto-send"]
            if "auto-delete" in submission:
                self.new_doc["auto_delete"] = submission["auto-delete"]

    def _cleanup_children(self):
        def remove_refs(children):
            for child in children:
                if isinstance(child, dict):
                    if "nodeset" in child:
                        del child["nodeset"]
                    if "ref" in child:
                        del child["ref"]
                    if "__order" in child:
                        del child["__order"]
                    if "children" in child:
                        remove_refs(child["children"])

        # do some ordering, order is specified by bindings
        def order_children(children):
            if isinstance(children, list):
                try:
                    children.sort(key=itemgetter("__order"))
                except KeyError:
                    pass
                for child in children:
                    if isinstance(child, dict) and "children" in child:
                        order_children(child["children"])

        order_children(self.children)
        remove_refs(self.children)

    def _cleanup_bind_list(self):
        for item in self._bind_list:
            ref = item["nodeset"]
            name = self._get_name_from_ref(ref)
            parent_ref = ref[: ref.find("/%s" % name)]
            question = self._get_question_params_from_bindings(ref)
            question["name"] = name
            question["__order"] = self._get_question_order(ref)
            if "calculate" in item:
                question["type"] = "calculate"
            if ref.split("/").__len__() == 3:
                # just append on root node, has no group
                question["ref"] = ref
                self.children.append(question)
                continue
            for child in self.children:
                if child["ref"] == parent_ref:
                    question["ref"] = ref
                    updated = False
                    for c in child["children"]:
                        if isinstance(c, dict) and "ref" in c and c["ref"] == ref:
                            c.update(question)
                            updated = True
                    if not updated:
                        child["children"].append(question)
            if "ref" not in question:
                new_ref = "/".join(ref.split("/")[2:])
                root_ref = "/".join(ref.split("/")[:2])
                q = self._get_item_func(root_ref, new_ref, item)
                if "type" not in q and "type" in question:
                    q.update(question)
                if q["type"] == "group" and q["name"] == "meta":
                    q["control"] = {"bodyless": True}
                    q["__order"] = self._get_question_order(ref)
                self.children.append(q)
                self._bind_list.append(item)
                break
        if self._bind_list:
            self._cleanup_bind_list()

    def _get_item_func(self, ref, name, item):
        rs = {}
        name_splits = name.split("/")
        rs["name"] = name_splits[0]
        ref = "%s/%s" % (ref, rs["name"])
        rs["ref"] = ref
        if name_splits.__len__() > 1:
            rs["type"] = "group"
            rs["children"] = [self._get_item_func(ref, "/".join(name_splits[1:]), item)]
        return rs

    def survey(self):
        new_doc = json.dumps(self.new_doc)
        _survey = builder.create_survey_element_from_json(new_doc)
        return _survey

    def _get_question_order(self, ref):
        try:
            return self.ordered_binding_refs.index(ref)
        except ValueError:
            # likely a group
            for i in self.ordered_binding_refs:
                if i.startswith(ref):
                    return self.ordered_binding_refs.index(i) + 1
            return self.ordered_binding_refs.__len__() + 1

    def _get_question_from_object(self, obj, type=None):
        try:
            ref = obj["ref"]
        except KeyError:
            try:
                ref = obj["nodeset"]
            except KeyError:
                raise TypeError(
                    'cannot find "ref" or "nodeset" in {}'.format(repr(obj))
                )
        question = {
            "ref": ref,
            "__order": self._get_question_order(ref),
            "name": self._get_name_from_ref(ref),
        }
        if "hint" in obj:
            k, v = self._get_label(obj["hint"], "hint")
            question[k] = v
        if "label" in obj:
            k, v = self._get_label(obj["label"])
            if isinstance(v, dict) and "label" in v.keys() and "media" in v.keys():
                for _k, _v in iter(v.items()):
                    question[_k] = _v
            else:
                question[k] = v
        if "autoplay" in obj or "appearance" in obj or "count" in obj or "rows" in obj:
            question["control"] = {}
        if "appearance" in obj:
            question["control"].update({"appearance": obj["appearance"]})
        if "rows" in obj:
            question["control"].update({"rows": obj["rows"]})
        if "autoplay" in obj:
            question["control"].update({"autoplay": obj["autoplay"]})
        question_params = self._get_question_params_from_bindings(ref)

        if isinstance(question_params, dict):
            for k, v in iter(question_params.items()):
                question[k] = v

        # Some values set from bindings are incorrect or incomplete. Correct them now.
        if "mediatype" in obj:
            question["type"] = obj["mediatype"].replace("/*", "")
        if "item" in obj:
            children = []
            for i in obj["item"]:
                if isinstance(i, dict) and "label" in i.keys() and "value" in i.keys():
                    k, v = self._get_label(i["label"])
                    children.append({"name": i["value"], k: v})
            question["children"] = children
        question_type = question["type"] if "type" in question else type
        if (
            question_type == "text"
            and "bind" in question
            and "readonly" in question["bind"]
        ):
            question_type = question["type"] = "note"
            del question["bind"]["readonly"]
            if len(question["bind"].keys()) == 0:
                del question["bind"]
        if question_type in ["group", "repeat"]:
            if question_type == "group" and "repeat" in obj:
                question["children"] = self._get_children_questions(obj["repeat"])
                question_type = "repeat"
                if "count" in obj["repeat"]:
                    if "control" not in question:
                        question["control"] = {}
                    question["control"].update(
                        {
                            "jr:count": self._shorten_xpaths_in_string(
                                obj["repeat"]["count"].strip()
                            )
                        }
                    )
            else:
                question["children"] = self._get_children_questions(obj)
            question["type"] = question_type
        if type == "trigger":
            question["type"] = "acknowledge"
        if type in [
            "select1",
            "select",
        ]:  # Select bind type is 'string' https://github.com/XLSForm/pyxform/issues/168
            question["type"] = self.QUESTION_TYPES[type]
        if question_type == "geopoint" and "hint" in question:
            del question["hint"]
        if "type" not in question and type:
            question["type"] = question_type
        return question

    def _get_children_questions(self, obj):
        children = []
        for k, v in iter(obj.items()):
            if k in ["ref", "label", "nodeset"]:
                continue
            if isinstance(v, dict):
                child = self._get_question_from_object(v, type=k)
                children.append(child)
            elif isinstance(v, list):
                for i in v:
                    child = self._get_question_from_object(i, type=k)
                    children.append(child)
        return children

    def _get_question_params_from_bindings(self, ref):
        for item in self.bindings:
            if item["nodeset"] == ref:
                try:
                    self._bind_list.remove(item)
                except ValueError:
                    pass
                rs = {}
                for k, v in iter(item.items()):
                    if k == "nodeset":
                        continue
                    if k == "type":
                        v = self._get_question_type(v)
                    if k in [
                        "relevant",
                        "required",
                        "constraint",
                        "constraintMsg",
                        "readonly",
                        "calculate",
                        "noAppErrorString",
                        "requiredMsg",
                    ]:
                        if k == "noAppErrorString":
                            k = "jr:noAppErrorString"
                        if k == "requiredMsg":
                            k = "jr:requiredMsg"
                        if k == "constraintMsg":
                            k = "jr:constraintMsg"
                            v = self._get_constraint_msg(v)
                        if k == "required":
                            if v == "true()":
                                v = "yes"
                            elif v == "false()":
                                v = "no"
                        if k in ["constraint", "relevant", "calculate"]:
                            v = self._shorten_xpaths_in_string(v)
                        if "bind" not in rs:
                            rs["bind"] = {}
                        rs["bind"][k] = v
                        continue
                    if k == "preload" and v == "uid":
                        if "bind" not in rs:
                            rs["bind"] = {}
                        rs["bind"]["jr:preload"] = v
                    rs[k] = v
                if "preloadParams" in rs and "preload" in rs:
                    rs["type"] = rs["preloadParams"]
                    del rs["preloadParams"]
                    del rs["preload"]
                return rs
        return None

    def _get_question_type(self, type):
        if type in self.QUESTION_TYPES.keys():
            return self.QUESTION_TYPES[type]
        return type

    def _set_translations(self):
        if "itext" not in self.model:
            self.translations = []
            return
        assert "translation" in self.model["itext"]
        self.translations = self.model["itext"]["translation"]
        if isinstance(self.translations, dict):
            self.translations = [self.translations]
        assert "text" in self.translations[0]
        assert "lang" in self.translations[0]

    def _get_label(self, label_obj, key="label"):
        if isinstance(label_obj, dict):
            try:
                ref = label_obj["ref"].replace("jr:itext('", "").replace("')", "")
            except KeyError:
                return key, self._get_output_text(label_obj)
            else:
                return self._get_text_from_translation(ref, key)
        return key, label_obj

    def _get_output_text(self, value):
        if "output" in value and "_text" in value:
            v = [value["_text"], self._get_bracketed_name(value["output"]["value"])]
            text = " ".join(v)
            if "tail" in value["output"]:
                text = "".join([text, value["output"]["tail"]])
        elif "output" in value and "_text" not in value:
            text = self._get_bracketed_name(value["output"]["value"])
        else:
            return value
        return text

    def _get_text_from_translation(self, ref, key="label"):
        label = {}
        for translation in self.translations:
            lang = translation["lang"]
            label_list = translation["text"]
            for l in label_list:
                if "value" not in l or l["value"] == "-":  # skip blank label
                    continue
                if l["id"] == ref:
                    text = value = l["value"]
                    if isinstance(value, dict):
                        if "output" in value:
                            text = self._get_output_text(value)
                        if "form" in value and "_text" in value:
                            key = "media"
                            v = value["_text"]
                            if value["form"] == "image":
                                v = v.replace("jr://images/", "")
                            else:
                                v = v.replace("jr://%s/" % value["form"], "")
                            if v == "-":  # skip blank
                                continue
                            text = {value["form"]: v}
                    if isinstance(value, list):
                        for item in value:
                            if "form" in item and "_text" in item:
                                k = "media"
                                m_type = item["form"]
                                v = item["_text"]
                                if m_type == "image":
                                    v = v.replace("jr://images/", "")
                                else:
                                    v = v.replace("jr://%s/" % m_type, "")
                                if v == "-":
                                    continue
                                if k not in label:
                                    label[k] = {}
                                if m_type not in label[k]:
                                    label[k][m_type] = {}
                                label[k][m_type][lang] = v
                                continue
                            if isinstance(item, basestring):
                                if item == "-":
                                    continue
                            if "label" not in label:
                                label["label"] = {}
                            label["label"][lang] = item
                        continue

                    label[lang] = text
                    break
        if key == "media" and label.keys() == ["default"]:
            label = label["default"]
        return key, label

    def _get_bracketed_name(self, ref):
        name = self._get_name_from_ref(ref)
        return "".join(["${", name.strip(), "}"])

    def _get_constraint_msg(self, constraint_msg):
        if isinstance(constraint_msg, basestring):
            if constraint_msg.find(":jr:constraintMsg") != -1:
                ref = constraint_msg.replace("jr:itext('", "").replace("')", "")
                k, constraint_msg = self._get_text_from_translation(ref)
        return constraint_msg

    def _get_name_from_ref(self, ref):
        """given /xlsform_spec_test/launch,
        return the string after the last occurance of the character '/'
        """
        pos = ref.rfind("/")
        if pos == -1:
            return ref
        else:
            return ref[pos + 1 :].strip()

    def _expand_child(self, obj_list):
        return obj_list

    def _shorten_xpaths_in_string(self, text):
        def get_last_item(xpath_str):
            last_item = xpath_str.split("/")
            return last_item[len(last_item) - 1].strip()

        def replace_function(match):
            return "${%s}" % get_last_item(match.group())

        # moving re flags into compile for python 2.6 compat
        pattern = "( /[a-z0-9-_]+(?:/[a-z0-9-_]+)+ )"
        text = re.compile(pattern, flags=re.I).sub(replace_function, text)
        pattern = "(/[a-z0-9-_]+(?:/[a-z0-9-_]+)+)"
        text = re.compile(pattern, flags=re.I).sub(replace_function, text)
        return text


def write_object_to_file(filename, obj):
    """Writes to a JSON filename the dictionary obj."""
    with codecs.open(filename, "w", encoding="utf-8") as output_file:
        output_file.write(json.dumps(obj, indent=2))
    logger.info("object written to file: %s", filename)



"""
Test upload (image, audio, file) question types in XLSForm
"""
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class UploadTest(PyxformTestCase):
    def test_image_question(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |      |         |       |
            |        | type | name    | label |
            |        | image | photo | Take a photo: |
            """,
            errored=False,
            xml__contains=[
                '<bind nodeset="/data/photo" type="binary"/>',
                '<upload mediatype="image/*" ref="/data/photo">',
                "<label>Take a photo:</label>",
                "</upload>",
            ],
        )

    def test_audio_question(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |      |         |       |
            |        | type | name    | label |
            |        | audio | recording1 | Record a sound: |
            """,
            errored=False,
            xml__contains=[
                '<bind nodeset="/data/recording1" type="binary"/>',
                '<upload mediatype="audio/*" ref="/data/recording1">',
                "<label>Record a sound:</label>",
                "</upload>",
            ],
        )

    def test_file_question(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |      |         |       |
            |        | type | name    | label |
            |        | file | file1 | Upload a file: |
            """,
            errored=False,
            xml__contains=[
                '<bind nodeset="/data/file1" type="binary"/>',
                '<upload mediatype="application/*" ref="/data/file1">',
                "<label>Upload a file:</label>",
                "</upload>",
            ],
        )

    def test_file_question_restrict_filetype(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |      |         |       |       |
            |        | type | name    | label | body::accept |
            |        | file | upload_a_pdf | Upload a PDF: | application/pdf |
            """,
            errored=False,
            xml__contains=['<upload accept="application/pdf"'],
        )

    def test_image_question_custom_col_calc(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |       |                  |                 |                               |
            |        | type  | name             | label           | body:esri:style               |
            |        | text  | watermark_phrase | Watermark Text: |                               |
            |        | text  | text1            | Text            |                               |
            |        | image | image1           | Take a Photo:   | watermark=${watermark_phrase} |
            """,  # noqa
            errored=False,
            xml__contains=["watermark= /data/watermark_phrase "],
        )



# -*- coding: utf-8 -*-
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class BackgroundAudioTest(PyxformTestCase):
    def test_background_audio(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |                  |              |
            |        | type             | name         |
            |        | background-audio | my_recording |
            """,
            xml__contains=[
                '<bind nodeset="/data/my_recording" type="binary"/>',
                '<odk:recordaudio event="odk-instance-load" ref="/data/my_recording"/>',
            ],
        )

    def test_background_audio_voice_only(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |                  |              |                      |
            |        | type             | name         | parameters           |
            |        | background-audio | my_recording | quality=voice-only |
            """,
            xml__contains=[
                '<odk:recordaudio event="odk-instance-load" ref="/data/my_recording" odk:quality="voice-only"/>',
            ],
        )

    def test_background_audio_low(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |                  |              |                      |
            |        | type             | name         | parameters           |
            |        | background-audio | my_recording | quality=low |
            """,
            xml__contains=[
                '<odk:recordaudio event="odk-instance-load" ref="/data/my_recording" odk:quality="low"/>',
            ],
        )

    def test_background_audio_normal(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |                  |              |                      |
            |        | type             | name         | parameters           |
            |        | background-audio | my_recording | quality=normal |
            """,
            xml__contains=[
                '<odk:recordaudio event="odk-instance-load" ref="/data/my_recording" odk:quality="normal"/>',
            ],
        )

    def test_external_quality_fails(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |                  |              |                      |
            |        | type             | name         | parameters           |
            |        | background-audio | my_recording | quality=external |
            """,
            errored=True,
            error__contains=["Invalid value for quality."],
        )

    def test_foo_quality_fails(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |                  |              |                      |
            |        | type             | name         | parameters           |
            |        | background-audio | my_recording | quality=foo |
            """,
            errored=True,
            error__contains=["Invalid value for quality."],
        )



# -*- coding: utf-8 -*-
"""
Test duplicate survey question field name.
"""
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class FieldsTests(PyxformTestCase):
    """
    Test XLSForm Fields
    """

    def test_duplicate_fields(self):
        """
        Ensure that duplicate field names are not allowed
        """
        self.assertPyxformXform(
            name="duplicatefields",
            md="""
            | Survey |         |         |               |
            |        | Type    | Name    | Label         |
            |        | integer | age     | the age       |
            |        | integer | age     | the age       |
            """,
            errored=True,
            error__contains=["There are more than one survey elements named 'age'"],
        )

    def test_duplicate_fields_diff_cases(self):
        """
        Ensure that duplicate field names with different cases are not allowed
        """
        self.assertPyxformXform(
            name="duplicatefieldsdiffcases",
            md="""
            | Survey |         |         |               |
            |        | Type    | Name    | Label         |
            |        | integer | age     | the age       |
            |        | integer | Age     | the age       |
            """,
            errored=True,
            error__contains=["There are more than one survey elements named 'age'"],
        )

    def test_duplicate_choices_without_setting(self):
        self.assertPyxformXform(
            md="""
            | survey  |                 |          |          |
            |         | type            | name     | label    |
            |         | select_one list | S1       | s1       |
            | choices |                 |          |          |
            |         | list name       | name     | label    |
            |         | list            | a        | option a |
            |         | list            | b        | option b |
            |         | list            | b        | option c |
            """,
            errored=True,
            error__contains=[
                "The name column for the 'list' choice list contains these duplicates: 'b'"
            ],  # noqa
        )

    def test_multiple_duplicate_choices_without_setting(self):
        self.assertPyxformXform(
            md="""
            | survey  |                 |          |          |
            |         | type            | name     | label    |
            |         | select_one list | S1       | s1       |
            | choices |                 |          |          |
            |         | list name       | name     | label    |
            |         | list            | a        | option a |
            |         | list            | a        | option b |
            |         | list            | b        | option c |
            |         | list            | b        | option d |
            """,
            errored=True,
            error__contains=[
                "The name column for the 'list' choice list contains these duplicates: 'a', 'b'"
            ],  # noqa
        )

    def test_duplicate_choices_with_setting_not_set_to_yes(self):
        self.assertPyxformXform(
            md="""
            | survey  |                 |          |          |
            |         | type            | name     | label    |
            |         | select_one list | S1       | s1       |
            | choices |                 |          |          |
            |         | list name       | name     | label    |
            |         | list            | a        | option a |
            |         | list            | b        | option b |
            |         | list            | b        | option c |
            | settings |                |          |          |
            |          | id_string    | allow_choice_duplicates   |
            |          | Duplicates   | Bob                       |
            """,
            errored=True,
            error__contains=[
                "The name column for the 'list' choice list contains these duplicates: 'b'"
            ],  # noqa
        )

    def test_duplicate_choices_with_allow_choice_duplicates_setting(self):
        md = """
            | survey  |                 |          |          |
            |         | type            | name     | label    |
            |         | select_one list | S1       | s1       |
            | choices |                 |          |          |
            |         | list name       | name     | label    |
            |         | list            | a        | option a |
            |         | list            | b        | option b |
            |         | list            | b        | option c |
            | settings |                |          |          |
            |          | id_string      | allow_choice_duplicates   |
            |          | Duplicates     | Yes                       |
            """

        expected = """
    <select1 ref="/pyxform_autotestname/S1">
      <label>s1</label>
      <item>
        <label>option a</label>
        <value>a</value>
      </item>
      <item>
        <label>option b</label>
        <value>b</value>
      </item>
      <item>
        <label>option c</label>
        <value>b</value>
      </item>
    </select1>
"""
        self.assertPyxformXform(md=md, xml__contains=[expected])

    def test_choice_list_without_duplicates_is_successful(self):
        md = """
            | survey  |                 |          |       |
            |         | type            | name     | label |
            |         | select_one list | S1       | s1    |
            | choices |                 |          |       |
            |         | list name       | name     | label  |
            |         | list            | option a | a      |
            |         | list            | option b | b      |
            | settings |                |          |        |
            |          | id_string    | allow_choice_duplicates   |
            |          | Duplicates   | Yes                       |
            """

        expected = """
    <select1 ref="/pyxform_autotestname/S1">
      <label>s1</label>
      <item>
        <label>a</label>
        <value>option a</value>
      </item>
      <item>
        <label>b</label>
        <value>option b</value>
      </item>
    </select1>
"""
        self.assertPyxformXform(md=md, xml__contains=[expected])

    def test_duplicate_form_name_in_section_name(self):
        """
        Ensure that the section name cannot be the same as form name
        """
        self.assertPyxformXform(
            name="foo",
            md="""
            | Survey   |             |         |               |
            |          | Type        | Name    | Label         |
            |          | begin group | foo     | A group       |
            |          | text        | a       | Enter text    |
            |          | end group   |         |               |
            """,
            errored=True,
            error__contains=['The name "foo" is the same as the form name'],
        )

    def test_field_name_may_match_form_name(self):
        """
        Unlike section names, it's okay for a field name to match the form
        name, which becomes the XML root node name
        """
        self.assertPyxformXform(
            name="activity",
            md="""
            | survey  |             |          |                   |
            |         | type        | name     | label             |
            |         | date        | date     | Observation date  |
            |         | text        | activity | Describe activity |
            """,
            errored=False,
        )



# -*- coding: utf-8 -*-
"""
Some tests for the new (v0.9) spec is properly implemented.
"""
import codecs
import os

import unittest2 as unittest

import pyxform
from pyxform.validators.odk_validate import check_xform, ODKValidateError
from pyxform.errors import PyXFormError
from pyxform.tests.utils import XFormTestCase
from pyxform.utils import has_external_choices
from pyxform.xls2json import SurveyReader, parse_file_to_workbook_dict
from pyxform.xls2json_backends import xls_to_dict

DIR = os.path.dirname(__file__)


class GroupNames(unittest.TestCase):
    maxDiff = None

    def runTest(self):
        filename = "group_name_test.xls"
        path_to_excel_file = os.path.join(DIR, "bug_example_xls", filename)
        # Get the xform output path:
        root_filename, ext = os.path.splitext(filename)
        output_path = os.path.join(DIR, "test_output", root_filename + ".xml")
        # Do the conversion:
        warnings = []
        with self.assertRaises(Exception):
            json_survey = pyxform.xls2json.parse_file_to_json(
                path_to_excel_file, default_name="group_name_test", warnings=warnings
            )
            survey = pyxform.create_survey_element_from_dict(json_survey)
            survey.print_xform_to_file(output_path, warnings=warnings)


class NotClosedGroup(unittest.TestCase):
    maxDiff = None

    def runTest(self):
        filename = "not_closed_group_test.xls"
        path_to_excel_file = os.path.join(DIR, "bug_example_xls", filename)
        # Get the xform output path:
        root_filename, ext = os.path.splitext(filename)
        output_path = os.path.join(DIR, "test_output", root_filename + ".xml")
        # Do the conversion:
        warnings = []
        with self.assertRaises(Exception):
            json_survey = pyxform.xls2json.parse_file_to_json(
                path_to_excel_file,
                default_name="not_closed_group_test",
                warnings=warnings,
            )
            survey = pyxform.create_survey_element_from_dict(json_survey)
            survey.print_xform_to_file(output_path, warnings=warnings)


class DuplicateColumns(unittest.TestCase):
    maxDiff = None

    def runTest(self):
        filename = "duplicate_columns.xlsx"
        path_to_excel_file = os.path.join(DIR, "example_xls", filename)
        # Get the xform output path:
        root_filename, ext = os.path.splitext(filename)
        output_path = os.path.join(DIR, "test_output", root_filename + ".xml")
        # Do the conversion:
        warnings = []
        with self.assertRaises(Exception):
            json_survey = pyxform.xls2json.parse_file_to_json(
                path_to_excel_file, default_name="duplicate_columns", warnings=warnings
            )
            survey = pyxform.create_survey_element_from_dict(json_survey)
            survey.print_xform_to_file(output_path, warnings=warnings)


class RepeatDateTest(XFormTestCase):
    maxDiff = None

    def runTest(self):
        filename = "repeat_date_test.xls"
        self.get_file_path(filename)
        expected_output_path = os.path.join(
            DIR, "test_expected_output", self.root_filename + ".xml"
        )

        # Do the conversion:
        warnings = []
        json_survey = pyxform.xls2json.parse_file_to_json(
            self.path_to_excel_file, default_name="repeat_date_test", warnings=warnings
        )
        survey = pyxform.create_survey_element_from_dict(json_survey)
        survey.print_xform_to_file(self.output_path, warnings=warnings)
        # print warnings
        # Compare with the expected output:
        with codecs.open(expected_output_path, "rb", encoding="utf-8") as expected_file:
            with codecs.open(self.output_path, "rb", encoding="utf-8") as actual_file:
                self.assertXFormEqual(expected_file.read(), actual_file.read())


class XmlEscaping(XFormTestCase):
    maxDiff = None

    def runTest(self):
        filename = "xml_escaping.xls"
        self.get_file_path(filename)
        expected_output_path = os.path.join(
            DIR, "test_expected_output", self.root_filename + ".xml"
        )

        # Do the conversion:
        warnings = []
        json_survey = pyxform.xls2json.parse_file_to_json(
            self.path_to_excel_file, default_name="xml_escaping", warnings=warnings
        )
        survey = pyxform.create_survey_element_from_dict(json_survey)
        survey.print_xform_to_file(self.output_path, warnings=warnings)
        # print warnings
        # Compare with the expected output:
        with codecs.open(expected_output_path, "rb", encoding="utf-8") as expected_file:
            with codecs.open(self.output_path, "rb", encoding="utf-8") as actual_file:
                self.assertXFormEqual(expected_file.read(), actual_file.read())


class DefaultTimeTest(XFormTestCase):
    maxDiff = None

    def runTest(self):
        filename = "default_time_demo.xls"
        path_to_excel_file = os.path.join(DIR, "bug_example_xls", filename)
        # Get the xform output path:
        root_filename, ext = os.path.splitext(filename)
        output_path = os.path.join(DIR, "test_output", root_filename + ".xml")
        expected_output_path = os.path.join(
            DIR, "test_expected_output", root_filename + ".xml"
        )
        # Do the conversion:
        warnings = []
        json_survey = pyxform.xls2json.parse_file_to_json(
            path_to_excel_file, default_name="default_time_demo", warnings=warnings
        )
        survey = pyxform.create_survey_element_from_dict(json_survey)
        survey.print_xform_to_file(output_path, warnings=warnings)
        # print warnings
        # Compare with the expected output:
        with codecs.open(expected_output_path, "rb", encoding="utf-8") as expected_file:
            with codecs.open(output_path, "rb", encoding="utf-8") as actual_file:
                self.assertXFormEqual(expected_file.read(), actual_file.read())


class ValidateWrapper(unittest.TestCase):
    maxDiff = None

    def runTest(self):
        filename = "ODKValidateWarnings.xlsx"
        path_to_excel_file = os.path.join(DIR, "bug_example_xls", filename)
        # Get the xform output path:
        root_filename, ext = os.path.splitext(filename)
        output_path = os.path.join(DIR, "test_output", root_filename + ".xml")
        # Do the conversion:
        warnings = []
        json_survey = pyxform.xls2json.parse_file_to_json(
            path_to_excel_file, default_name="ODKValidateWarnings", warnings=warnings
        )
        survey = pyxform.create_survey_element_from_dict(json_survey)
        survey.print_xform_to_file(output_path, warnings=warnings)


class EmptyStringOnRelevantColumnTest(unittest.TestCase):
    def runTest(self):
        filename = "ict_survey_fails.xls"
        path_to_excel_file = os.path.join(DIR, "bug_example_xls", filename)
        workbook_dict = pyxform.xls2json.parse_file_to_workbook_dict(path_to_excel_file)
        with self.assertRaises(KeyError):
            # bind:relevant should not be part of workbook_dict
            workbook_dict["survey"][0]["bind: relevant"].strip()


class BadChoicesSheetHeaders(unittest.TestCase):
    def runTest(self):
        filename = "spaces_in_choices_header.xls"
        path_to_excel_file = os.path.join(DIR, "bug_example_xls", filename)
        warnings = []
        pyxform.xls2json.parse_file_to_json(
            path_to_excel_file,
            default_name="spaces_in_choices_header",
            warnings=warnings,
        )
        self.assertEquals(len(warnings), 3, "Found " + str(len(warnings)) + " warnings")

    def test_values_with_spaces_are_cleaned(self):
        """
        Test that values with leading and trailing whitespaces are processed.
    
        This test checks that the submission_url provided is cleaned
        of leading and trailing whitespaces.
        """
        filename = "spaces_in_choices_header.xls"
        path_to_excel_file = os.path.join(DIR, "bug_example_xls", filename)
        survey_reader = SurveyReader(
            path_to_excel_file, default_name="spaces_in_choices_header"
        )
        result = survey_reader.to_json_dict()

        self.assertEqual(
            result["submission_url"], "https://odk.ona.io/random_person/submission"
        )


class TestChoiceNameAsType(unittest.TestCase):
    def test_choice_name_as_type(self):
        filename = "choice_name_as_type.xls"
        path_to_excel_file = os.path.join(DIR, "example_xls", filename)
        xls_reader = SurveyReader(
            path_to_excel_file, default_name="choice_name_as_type"
        )
        survey_dict = xls_reader.to_json_dict()
        self.assertTrue(has_external_choices(survey_dict))


class TestBlankSecondRow(unittest.TestCase):
    def test_blank_second_row(self):
        filename = "blank_second_row.xls"
        path_to_excel_file = os.path.join(DIR, "bug_example_xls", filename)
        xls_reader = SurveyReader(path_to_excel_file, default_name="blank_second_row")
        survey_dict = xls_reader.to_json_dict()
        self.assertTrue(len(survey_dict) > 0)


class TestXLDateAmbigous(unittest.TestCase):
    """Test non standard sheet with exception is processed successfully."""

    def test_xl_date_ambigous(self):
        """Test non standard sheet with exception is processed successfully."""
        filename = "xl_date_ambiguous.xlsx"
        path_to_excel_file = os.path.join(DIR, "bug_example_xls", filename)
        xls_reader = SurveyReader(path_to_excel_file, default_name="xl_date_ambiguous")
        survey_dict = xls_reader.to_json_dict()
        self.assertTrue(len(survey_dict) > 0)


class TestXLDateAmbigousWithException(unittest.TestCase):
    """Test non standard sheet date values to raise an exception.
        This exception is raised if the date values exceed the
        datemode value accepted by that workbook."""

    def test_xl_date_ambigous_with_exception(self):
        """Test non standard sheet with exception is processed successfully."""
        filename = "xl_date_ambiguous_v1.xlsx"
        path_to_excel_file = os.path.join(DIR, "bug_example_xls", filename)
        with self.assertRaises(PyXFormError) as e:
            xls_to_dict(path_to_excel_file)
        msg = (
            "The xls file provided has an invalid date on the"
            " survey sheet, under the default column on row number 5"
        )

        self.assertEqual(msg, str(e.exception))


class TestSpreadSheetFilesWithMacrosAreAllowed(unittest.TestCase):
    """Test that spreadsheets with .xlsm extension are allowed"""

    def test_xlsm_files_are_allowed(self):
        filename = "excel_with_macros.xlsm"
        path_to_excel_file = os.path.join(DIR, "bug_example_xls", filename)
        result = parse_file_to_workbook_dict(path_to_excel_file)
        self.assertIsInstance(result, dict)


class TestBadCalculation(unittest.TestCase):
    """Bad calculation should not kill the application"""

    def test_bad_calculate_javarosa_error(self):
        filename = "bad_calc.xml"
        test_xml = os.path.join(DIR, "test_output", filename)
        self.assertRaises(ODKValidateError, check_xform, test_xml)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
BindConversionsTest - test bind conversions.
"""
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class BindConversionsTest(PyxformTestCase):
    """
    BindConversionsTest - test bind conversions
    """

    def test_bind_readonly_conversion(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |       |      |       |          |
            |        | type  | name | label | readonly |
            |        | text  | text | text  | yes      |
            """,
            xml__contains=['<bind nodeset="/data/text" readonly="true()"'],
        )

    def test_bind_required_conversion(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |       |      |       |          |
            |        | type  | name | label | required |
            |        | text  | text | text  | FALSE    |
            """,
            xml__contains=['<bind nodeset="/data/text" required="false()"'],
        )

    def test_bind_required_message_with_reference(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |       |      |       |            |                    |
            |        | type  | name | label | required   | required_message |
            |        | int   | foo  | foo   |            |                    |
            |        | text  | text | text  | true()     | required, ${foo}   |
            """,
            xml__contains=[
                '<bind nodeset="/data/text" required="true()" type="string" jr:requiredMsg="jr:itext(\'/data/text:jr:requiredMsg\')"',
                '<value> required, <output value=" /data/foo "/> </value>',
            ],
        )

    def test_bind_constraint_conversion(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |       |      |       |                    |
            |        | type  | name | label | constraint_message |
            |        | text  | text | text  | yes                |
            """,
            xml__contains=[
                '<bind nodeset="/data/text" type="string" jr:constraintMsg="yes"'
            ],
        )

    def test_bind_constraint_message_with_reference(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |       |      |       |                      |                    |
            |        | type  | name | label | constraint           | constraint_message |
            |        | int   | foo  | foo   |                      |                    |
            |        | text  | text | text  | string-length(.) > 1 | too short ${foo}   |
            """,
            xml__contains=[
                '<bind constraint="string-length(.) &gt; 1" nodeset="/data/text" type="string" jr:constraintMsg="jr:itext(\'/data/text:jr:constraintMsg\')"',
                '<value> too short <output value=" /data/foo "/> </value>',
            ],
        )

    def test_bind_custom_conversion(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |       |      |       |           |
            |        | type  | name | label | bind::foo |
            |        | text  | text | text  | bar       |
            """,
            xml__contains=['<bind foo="bar" nodeset="/data/text" type="string"'],
        )



# -*- coding: utf-8 -*-
"""
Test xls2json_backends module functionality.
"""
from unittest import TestCase

import xlrd

from pyxform.xls2json_backends import xls_value_to_unicode


class TestXLS2JSONBackends(TestCase):
    """
    Test xls2json_backends module.
    """

    def test_xls_value_to_unicode(self):
        """
        Test external choices sheet with numeric values is processed successfully.

        The test ensures that the integer values within the external choices sheet
        are returned as they were initially received.
        """
        value = 32.0
        value_type = xlrd.XL_CELL_NUMBER
        datemode = 1
        csv_data = xls_value_to_unicode(value, value_type, datemode)
        expected_output = "32"
        self.assertEqual(csv_data, expected_output)

        # Test that the decimal value is not changed during conversion.
        value = 46.9
        csv_data = xls_value_to_unicode(value, value_type, datemode)
        expected_output = "46.9"
        self.assertEqual(csv_data, expected_output)



# -*- coding: utf-8 -*-
"""
pyxform_validator_update - command to update XForm validators.
"""
from __future__ import print_function

import argparse
import fnmatch
import io
import json
import logging
import os
import shutil
import sys
from datetime import datetime
from stat import S_IXGRP, S_IXUSR
from zipfile import ZipFile, is_zipfile

from pyxform.errors import PyXFormError
from pyxform.utils import unicode
from pyxform.validators import enketo_validate, odk_validate
from pyxform.validators.util import HERE, CapturingHandler, request_get

UTC_FMT = "%Y-%m-%dT%H:%M:%SZ"
log = logging.getLogger(__name__)
capture_handler = CapturingHandler(logger=log)


class _UpdateInfo(object):
    """
    Data class for Updater info.
    """

    def __init__(
        self,
        api_url,
        repo_url,
        validate_subfolder,
        install_check,
        validator_basename,
        mod_root=None,
    ):
        """
        :param api_url: The GitHub API URL for the latest release details.
        :param repo_url: The main GitHub repository page.
        :param validate_subfolder: The folder under "validators" to work in.
        :param install_check: A function to check if an install works. Must
            return True or False.
        :param validator_basename: Validator bin file name.
        :param mod_root: Optionally specify the root module path.
        """
        self._api_url = None
        self.api_url = api_url
        self.repo_url = repo_url
        self.validate_subfolder = validate_subfolder
        self.install_check = install_check
        self.validator_basename = validator_basename

        if mod_root is None:
            self.mod_path = os.path.join(HERE, self.validate_subfolder)
        else:
            self.mod_path = os.path.join(mod_root, validate_subfolder)
        self.latest_path = os.path.join(self.mod_path, "latest.json")
        self.last_check_path = os.path.join(self.mod_path, ".last_check")

        self.bin_path = os.path.join(self.mod_path, "bin")
        self.installed_path = os.path.join(self.bin_path, "installed.json")
        self.bin_new_path = os.path.join(self.mod_path, "bin_new")

        self.manual_msg = "Download manually from: {r}.".format(r=self.repo_url)


class _UpdateHandler(object):
    """
    Handles tasks related to updating external XForm validators.

    Where possible, minimise calls to the GitHub API because doing so without an
    API token is rate-limited to 60 calls per hour, tracked by IP address. The
    calls are unauthenticated to avoid having to handle / manage GitHub creds.
    Tests should not touch GitHub either.
    """

    @staticmethod
    def _request_latest_json(url):
        """
        Get the GitHub API JSON response doc for the latest release from URL.
        """
        content = request_get(url=url)
        return json.loads(content.decode("utf-8"))

    @staticmethod
    def _check_path(file_path):
        if not os.path.exists(file_path):
            raise PyXFormError(
                "Expected path does not exist: {p}" "".format(p=file_path)
            )
        else:
            return True

    @staticmethod
    def _read_json(file_path):
        """
        Read the JSON file to a string.
        """
        _UpdateHandler._check_path(file_path=file_path)
        with io.open(file_path, mode="r") as in_file:
            return json.load(in_file)

    @staticmethod
    def _write_json(file_path, content):
        """
        Save the JSON data to a file.
        """
        with io.open(file_path, mode="w", newline="\n") as out_file:
            data = json.dumps(content, indent=2, sort_keys=True)
            out_file.write(unicode(data))

    @staticmethod
    def _read_last_check(file_path):
        """
        Read the .last_check file.
        """
        _UpdateHandler._check_path(file_path=file_path)
        with io.open(file_path, mode="r") as in_file:
            first_line = in_file.readline()
        try:
            last_check = datetime.strptime(first_line, UTC_FMT)
        except ValueError:
            return None
        else:
            return last_check

    @staticmethod
    def _write_last_check(file_path, content):
        """
        Write the .last_check file.
        """
        with io.open(file_path, mode="w", newline="\n") as out_file:
            out_file.write(unicode(content.strftime(UTC_FMT)))

    @staticmethod
    def _check_necessary(update_info, utc_now):
        """
        Determine whether a check for the latest version is necessary.
        """
        if not os.path.exists(update_info.last_check_path):
            return True
        elif not os.path.exists(update_info.latest_path):
            return True
        else:
            last_check = _UpdateHandler._read_last_check(
                file_path=update_info.last_check_path
            )
            if last_check is None:
                return True
            age = utc_now - last_check
            thirty_minutes = 1800
            if thirty_minutes < age.total_seconds():
                return True
            else:
                return False

    @staticmethod
    def _get_latest(update_info):
        """
        Get the latest release info, either from GitHub or a recent file copy.
        """
        utc_now = datetime.utcnow()
        if _UpdateHandler._check_necessary(update_info=update_info, utc_now=utc_now):
            latest = _UpdateHandler._request_latest_json(url=update_info.api_url)
            _UpdateHandler._write_json(
                file_path=update_info.latest_path, content=latest
            )
            _UpdateHandler._write_last_check(
                file_path=update_info.last_check_path, content=utc_now
            )
        else:
            latest = _UpdateHandler._read_json(file_path=update_info.latest_path)
        return latest

    @staticmethod
    def _get_release_message(json_data):
        template = "- Tag name = {tag_name}\n" "- Tag URL = {tag_url}\n\n"
        return template.format(
            tag_name=json_data["tag_name"], tag_url=json_data["html_url"]
        )

    @staticmethod
    def list(update_info):
        """
        List the current and latest release info, and latest available files.

        :type update_info: _UpdateInfo
        """
        if not os.path.exists(update_info.installed_path):
            installed_info = "- None!\n\n"
        else:
            installed = _UpdateHandler._read_json(file_path=update_info.installed_path)
            installed_info = _UpdateHandler._get_release_message(json_data=installed)

        latest = _UpdateHandler._get_latest(update_info=update_info)
        latest_files = latest["assets"]
        if len(latest_files) == 0:
            file_message = "- None!\n\n{m}".format(m=update_info.manual_msg)
        else:
            file_names = ["- {n}".format(n=x["name"]) for x in latest_files]
            file_message = "\n".join(file_names)

        template = (
            "\nInstalled release:\n\n{installed}"
            "Latest release:\n\n{latest}"
            "Files available:\n\n{file_message}\n"
        )
        message = template.format(
            installed=installed_info,
            latest=_UpdateHandler._get_release_message(json_data=latest),
            file_message=file_message,
        )
        log.info(message)

    @staticmethod
    def _find_download_url(update_info, json_data, file_name):
        """
        Find the download URL for the file in the GitHub API JSON response doc.
        """
        rel_name = json_data["tag_name"]
        files = json_data["assets"]

        if len(files) == 0:
            raise PyXFormError(
                "No files attached to release '{r}'.\n\n{h}"
                "".format(r=rel_name, h=update_info.manual_msg)
            )

        file_urls = [x["browser_download_url"] for x in files if x["name"] == file_name]

        urls_len = len(file_urls)
        if 0 == urls_len:
            raise PyXFormError(
                "No files with the name '{n}' attached to release '{r}'."
                "\n\n{h}".format(n=file_name, r=rel_name, h=update_info.manual_msg)
            )
        elif 1 < urls_len:
            raise PyXFormError(
                "{c} files with the name '{n}' attached to release '{r}'."
                "\n\n{h}".format(
                    c=urls_len, n=file_name, r=rel_name, h=update_info.manual_msg
                )
            )
        else:
            return file_urls[0]

    @staticmethod
    def _download_file(url, file_path):
        """
        Save response content from the URL to a binary file at the file path.
        """
        with io.open(file_path, mode="wb") as out_file:
            file_data = request_get(url=url)
            out_file.write(file_data)

    @staticmethod
    def _get_bin_paths(update_info, file_path):
        """
        Get the mapping of zip file paths to extract paths for the file_name.

        The zip file paths are actually glob/fnmatch patterns to find these
        files among the files in the zip archive.
        """
        _, file_name = os.path.split(file_path)
        file_base = os.path.basename(file_name)
        if "windows" in file_base:
            main_bin = "*validate.exe"
        elif "linux" in file_base:
            main_bin = "*validate"
        elif "macos" in file_base:
            main_bin = "*validate"
        else:
            raise PyXFormError(
                "Did not find a supported main binary for file: {p}.\n\n{h}"
                "".format(p=file_path, h=update_info.manual_msg)
            )
        return [
            (main_bin, update_info.validator_basename),
            (
                "*node_modules/libxmljs-mt/build*/xmljs.node",
                "node_modules/libxmljs-mt/build/xmljs.node",
            ),
            (
                "*node_modules/libxslt/build*/node-libxslt.node",
                "node_modules/libxslt/build/node-libxslt.node",
            ),
        ]

    @staticmethod
    def _unzip_find_jobs(open_zip_file, bin_paths, out_path):
        """
        For each bin file, get the zip file item file name and the output path.

        Ignore files that may appear in the __MACOSX info dir, and if two files
        have the same destination path and the same CRC then they're probably
        duplicate files so only one of them is copied out.
        """
        zip_info = open_zip_file.infolist()
        zip_jobs = {}
        for zip_item in zip_info:
            if zip_item.filename.startswith("__MACOSX"):
                continue
            for file_target in bin_paths:
                if fnmatch.fnmatch(zip_item.filename, file_target[0]):
                    file_out_path = os.path.join(out_path, file_target[1])
                    maybe_existing_match = zip_jobs.get(file_out_path, None)
                    if maybe_existing_match is not None:
                        if maybe_existing_match.CRC == zip_item.CRC:
                            continue
                    zip_jobs[file_out_path] = zip_item
        if len(bin_paths) != len(zip_jobs.keys()):
            raise PyXFormError(
                "Expected {e} zip job files, found: {c}"
                "".format(e=len(bin_paths), c=len(zip_jobs.keys()))
            )
        return zip_jobs

    @staticmethod
    def _unzip_extract_file(open_zip_file, zip_item, file_out_path):
        """
        Extract the content for item in zip file to a specific location.

        Note that ZipExtFile.read() (which is returned from open()) does a
        CRC32 check during read so ZipFile.testzip() isn't needed. Testzip
        works by reading all the files in the archive, doing the same thing.
        """
        out_parent = os.path.dirname(file_out_path)
        if not os.path.exists(out_parent):
            os.makedirs(out_parent)
        with open_zip_file.open(zip_item, mode="r") as zip_item_file:
            zip_item_data = zip_item_file.read()
            with io.open(file_out_path, "wb") as file_out_file:
                file_out_file.write(zip_item_data)

    @staticmethod
    def _unzip(update_info, file_path, out_path):
        """
        Unzip the contents of a zip file to an existing output path.
        """
        _UpdateHandler._check_path(file_path=file_path)
        _UpdateHandler._check_path(file_path=out_path)
        bin_paths = _UpdateHandler._get_bin_paths(
            update_info=update_info, file_path=file_path
        )

        with ZipFile(file_path, mode="r") as zip_file:
            jobs = _UpdateHandler._unzip_find_jobs(
                open_zip_file=zip_file, bin_paths=bin_paths, out_path=out_path
            )
            for file_out_path, zip_item in jobs.items():
                _UpdateHandler._unzip_extract_file(
                    open_zip_file=zip_file,
                    zip_item=zip_item,
                    file_out_path=file_out_path,
                )

    @staticmethod
    def _install(update_info, file_name):
        """
        Install the latest release.
        """
        try:
            latest = _UpdateHandler._get_latest(update_info=update_info)
            file_path = os.path.join(update_info.bin_new_path, file_name)
            new_bin_file_path = os.path.join(
                update_info.bin_new_path, update_info.validator_basename
            )

            if os.path.exists(update_info.bin_new_path):
                shutil.rmtree(update_info.bin_new_path)
            os.makedirs(update_info.bin_new_path)

            installed = os.path.join(update_info.bin_new_path, "installed.json")
            _UpdateHandler._write_json(file_path=installed, content=latest)
            url = _UpdateHandler._find_download_url(
                update_info=update_info, json_data=latest, file_name=file_name
            )
            _UpdateHandler._download_file(url=url, file_path=file_path)

            if is_zipfile(file_path) and os.path.splitext(file_path)[1] == ".zip":
                _UpdateHandler._unzip(
                    update_info=update_info,
                    file_path=file_path,
                    out_path=update_info.bin_new_path,
                )
            else:
                os.rename(file_path, new_bin_file_path)

            # For macos / linux: chmod ug+x the new bin file. No-op on Windows.
            current_mode = os.stat(new_bin_file_path).st_mode
            os.chmod(new_bin_file_path, current_mode | S_IXUSR | S_IXGRP)

        except PyXFormError as e:
            raise PyXFormError("\n\nUpdate failed!\n\n" + unicode(e))
        else:
            return latest

    @staticmethod
    def _replace_old_bin_path(update_info):
        if os.path.exists(update_info.bin_path):
            shutil.rmtree(update_info.bin_path)
        shutil.move(update_info.bin_new_path, update_info.bin_path)

    @staticmethod
    def update(update_info, file_name, force=False):
        """
        Update to the latest version, using the specified file_name.

        :type update_info: _UpdateInfo
        :type file_name: str
        :type force: bool
        """
        if not os.path.exists(update_info.installed_path):
            installed = _UpdateHandler._install(
                update_info=update_info, file_name=file_name
            )
            latest = installed
        else:
            installed = _UpdateHandler._read_json(file_path=update_info.installed_path)
            latest = _UpdateHandler._get_latest(update_info=update_info)
            if installed["tag_name"] == latest["tag_name"] and not force:
                installed_info = _UpdateHandler._get_release_message(
                    json_data=installed
                )
                latest_info = _UpdateHandler._get_release_message(json_data=latest)
                template = (
                    "\nUpdate failed!\n\n"
                    "The installed release appears to be the latest. "
                    "To update anyway, use the '--force' flag.\n\n"
                    "Installed release:\n\n{installed}"
                    "Latest release:\n\n{latest}"
                )
                message = template.format(installed=installed_info, latest=latest_info)
                raise PyXFormError(message)
            else:
                _UpdateHandler._install(update_info=update_info, file_name=file_name)

        installed_info = _UpdateHandler._get_release_message(json_data=installed)
        latest_info = _UpdateHandler._get_release_message(json_data=latest)
        new_bin_file_path = os.path.join(
            update_info.bin_new_path, update_info.validator_basename
        )
        if update_info.install_check(bin_file_path=new_bin_file_path):
            _UpdateHandler._replace_old_bin_path(update_info=update_info)
            template = (
                "\nUpdate success!\n\n"
                "Install check of the latest release succeeded.\n\n"
                "Latest release:\n\n{latest}"
            )
            message = template.format(latest=latest_info)
            log.info(message)
            return True
        else:
            template = (
                "\nUpdate failed!\n\n"
                "The latest release does not appear to work. "
                "It is saved here in case it's needed:\n{bin_new}\n\n"
                "The installed release has not been changed.\n\n"
                "Installed release:\n\n{installed}"
                "Latest release:\n\n{latest}"
            )
            message = template.format(
                bin_new=new_bin_file_path, installed=installed_info, latest=latest_info
            )
            raise PyXFormError(message)

    @staticmethod
    def check(update_info):
        """
        Check if the installed release of the validator works.

        :type update_info: _UpdateInfo
        """
        if not os.path.exists(update_info.installed_path):
            message = "\nCheck failed!\n\n" "No installed release found."
            raise PyXFormError(message)

        installed = _UpdateHandler._read_json(file_path=update_info.installed_path)
        if update_info.install_check():
            template = (
                "\nCheck success!\n\n"
                "The installed release appears to work.\n\n"
                "Installed release:\n\n{installed}"
            )
            message = template.format(
                installed=_UpdateHandler._get_release_message(json_data=installed)
            )
            log.info(message)
            return True
        else:
            template = (
                "\nCheck failed!\n\n"
                "The installed release does not appear to work.\n\n"
                "Installed release:\n\n{installed}"
            )
            message = template.format(
                installed=_UpdateHandler._get_release_message(json_data=installed)
            )
            raise PyXFormError(message)


class _UpdateService(object):

    update_info = None

    def list(self):
        return _UpdateHandler.list(update_info=self.update_info)

    def update(self, file_name, force):
        return _UpdateHandler.update(
            update_info=self.update_info, file_name=file_name, force=force
        )

    def check(self):
        return _UpdateHandler.check(update_info=self.update_info)

    @staticmethod
    def _install_check(bin_file_path=None):
        raise NotImplementedError()


class EnketoValidateUpdater(_UpdateService):
    def __init__(self):
        self.update_info = _UpdateInfo(
            api_url="https://api.github.com/repos/enketo/enketo-validate/"
            "releases/latest",
            repo_url="https://github.com/enketo/enketo-validate",
            validate_subfolder="enketo_validate",
            install_check=self._install_check,
            validator_basename=os.path.basename(enketo_validate.ENKETO_VALIDATE_PATH),
        )

    @staticmethod
    def _install_check(bin_file_path=None):
        if bin_file_path is None:
            return enketo_validate.install_ok()
        else:
            extracted = os.path.join(os.path.dirname(bin_file_path), "validate")
            return enketo_validate.install_ok(bin_file_path=extracted)


class ODKValidateUpdater(_UpdateService):
    def __init__(self):
        self.update_info = _UpdateInfo(
            api_url="https://api.github.com/repos/opendatakit/validate/"
            "releases/latest",
            repo_url="https://github.com/opendatakit/validate",
            validate_subfolder="odk_validate",
            install_check=self._install_check,
            validator_basename=os.path.basename(odk_validate.ODK_VALIDATE_PATH),
        )

    @staticmethod
    def _install_check(bin_file_path=None):
        if bin_file_path is None:
            return odk_validate.install_ok()
        else:
            return odk_validate.install_ok(bin_file_path=bin_file_path)


def _build_validator_menu(main_subparser, validator_name, updater_instance):

    main = main_subparser.add_parser(
        validator_name.lower(),
        description="{v} Sub-menu".format(v=validator_name),
        help="For help, use '{v} -h'.".format(v=validator_name.lower()),
    )
    subs = main.add_subparsers(metavar="<sub_command>")

    cmd_list = subs.add_parser(
        "list", help="List available files for the latest release."
    )
    cmd_list.set_defaults(command=updater_instance.list)

    cmd_update = subs.add_parser(
        "update", help="Update the validator to the latest release."
    )
    cmd_update.set_defaults(command=updater_instance.update)
    cmd_update.add_argument(
        "file_name", help="Name of the release file to use for updating."
    )
    cmd_update.add_argument(
        "--force",
        action="store_true",
        default=False,
        help="If the installed release appears to be the latest, update anyway.",
    )

    cmd_check = subs.add_parser(
        "check", help="Check if the installed release of the validator appears to work."
    )
    cmd_check.set_defaults(command=updater_instance.check)
    return main


def _create_parser():
    """
    Parse command line arguments.
    """
    main_title = "pyxform validator updater"
    epilog = (
        "------------------------------------------------------\n"
        "Use this tool to update external validators.\n\n"
        "Example usage:\n\n"
        "updater.py enketo list\n"
        "updater.py enketo update linux.zip\n\n"
        "First, use the 'list' sub-command for the validator\n"
        "to check for a new release and to show what (if any) \n"
        "files are attached to it.\n\n"
        "Second, use the 'update' sub-command for the validator\n"
        "to apply the update, specifying the file to use.\n"
        "------------------------------------------------------"
    )
    main_parser = argparse.ArgumentParser(
        description=main_title,
        epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    sub_parsers = main_parser.add_subparsers(metavar="<sub_menu>")
    _build_validator_menu(
        main_subparser=sub_parsers,
        validator_name="Enketo",
        updater_instance=EnketoValidateUpdater(),
    )
    _build_validator_menu(
        main_subparser=sub_parsers,
        validator_name="ODK",
        updater_instance=ODKValidateUpdater(),
    )
    return main_parser


def main_cli():
    logger = logging.getLogger(name="pyxform_validator_update")
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(logging.INFO)

    try:
        parser = _create_parser()
        args = parser.parse_args()
        kwargs = args.__dict__.copy()
        del kwargs["command"]
        args.command(**kwargs)
    except PyXFormError as main_error:
        logger.info(unicode(main_error))
        sys.exit(1)
    if 0 < len(capture_handler.watcher.records):
        for line in capture_handler.watcher.output["INFO"]:
            logger.info(line)


if __name__ == "__main__":
    main_cli()



# -*- coding: utf-8 -*-
"""
Test validators.
"""

import sys
from unittest import TestCase
from pyxform.validators.odk_validate import check_java_available, _call_validator


if sys.version_info >= (3, 3):
    from unittest.mock import patch  # pylint: disable=E0611,E0401
else:
    try:
        from mock import patch
    except ImportError:
        raise ImportError("Pyxform test suite requires the 'mock' library.")


mock_func = "shutil.which"
msg = "Form validation failed because Java (8+ required) could not be found."


class TestValidatorsUtil(TestCase):
    """Test validators.util"""

    def test_check_java_available__found(self):
        """Should not raise an error when Java is found."""
        with patch(mock_func) as mocked:
            mocked.return_value = "/usr/bin/java"
            check_java_available()

    def test_check_java_available__not_found(self):
        """Should raise an error when java is not found."""
        with patch(mock_func) as mocked:
            mocked.return_value = None
            with self.assertRaises(EnvironmentError) as error:
                check_java_available()
            self.assertIn(msg, str(error.exception))



# -*- coding: utf-8 -*-
"""
Test rank widget.
"""
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class RangeWidgetTest(PyxformTestCase):
    def test_rank(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |              |          |       |
            |        | type         | name     | label |
            |        | rank mylist  | order    | Rank  |
            | choices|              |          |       |
            |        | list_name    | name     | label |
            |        | mylist       | a        | A     |
            |        | mylist       | b        | B     |
            """,
            xml__contains=[
                'xmlns:odk="http://www.opendatakit.org/xforms"',
                '<bind nodeset="/data/order" type="odk:rank"/>',
                '<odk:rank ref="/data/order">',
                "<label>Rank</label>",
                "<label>A</label>",
                "<value>a</value>",
                "<label>B</label>",
                "<value>b</value>",
                "</odk:rank>",
            ],
        )

    def test_rank_filter(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |              |          |       |               |
            |        | type         | name     | label | choice_filter |
            |        | rank mylist  | order    | Rank  | color='blue'  |
            | choices|              |          |       |
            |        | list_name    | name     | label | color |
            |        | mylist       | a        | A     | red   |
            |        | mylist       | b        | B     | blue  |
           """,
            xml__contains=[
                'xmlns:odk="http://www.opendatakit.org/xforms"',
                '<bind nodeset="/data/order" type="odk:rank"/>',
                '<instance id="mylist">',
                "<color>red</color>",
                "<name>a</name>",
                "<color>blue</color>",
                "<name>b</name>",
                """<odk:rank ref="/data/order">
      <label>Rank</label>
      <itemset nodeset="instance('mylist')/root/item[color='blue']">
        <value ref="name"/>
        <label ref="label"/>
      </itemset>
    </odk:rank>""",
            ],
        )

    def test_rank_translations(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |              |          |       |                    |
            |        | type         | name     | label | label::French (fr) |
            |        | rank mylist  | order    | Rank  | Ranger             |
            | choices|              |          |       |
            |        | list_name    | name     | label | label::French (fr) |
            |        | mylist       | a        | A     |  AA                |
            |        | mylist       | b        | B     |  BB                |
            """,
            xml__contains=[
                'xmlns:odk="http://www.opendatakit.org/xforms"',
                '<translation lang="French (fr)">',
                """<text id="/data/order:label">
            <value>Ranger</value>
          </text>""",
                """<text id="/data/order/a:label">
            <value>AA</value>
          </text>""",
                """<text id="/data/order/b:label">
            <value>BB</value>
          </text>""",
                "</translation>",
                """<odk:rank ref="/data/order">
      <label ref="jr:itext('/data/order:label')"/>
      <item>
        <label ref="jr:itext('/data/order/a:label')"/>
        <value>a</value>
      </item>
      <item>
        <label ref="jr:itext('/data/order/b:label')"/>
        <value>b</value>
      </item>
    </odk:rank>""",
            ],
        )



# -*- coding: utf-8 -*-
"""
Test language warnings.
"""
import os
import tempfile

from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class MetadataTest(PyxformTestCase):
    """
    Test metadata and related warnings.
    """

    def test_metadata_bindings(self):
        self.assertPyxformXform(
            name="metadata",
            md="""
            | survey |             |             |       |
            |        | type        | name        | label |
            |        | deviceid    | deviceid    |       |
            |        | phonenumber | phonenumber |       |
            |        | start       | start       |       |
            |        | end         | end         |       |
            |        | today       | today       |       |
            |        | username    | username    |       |
            |        | email       | email       |       |
            """,
            xml__contains=[
                'jr:preload="property" jr:preloadParams="deviceid"',
                'jr:preload="property" jr:preloadParams="phonenumber"',
                'jr:preload="timestamp" jr:preloadParams="start"',
                'jr:preload="timestamp" jr:preloadParams="end"',
                'jr:preload="date" jr:preloadParams="today"',
                'jr:preload="property" jr:preloadParams="username"',
                'jr:preload="property" jr:preloadParams="email"',
            ],
        )

    def test_simserial_deprecation_warning(self):
        warnings = []
        survey = self.md_to_pyxform_survey(
            """
            | survey |              |                          |                                            |
            |        | type         | name                     | label                                      |
            |        | simserial    | simserial                |                                            |
            |        | note         | simserial_test_output    | simserial_test_output: ${simserial}        |
            """,
            warnings=warnings,
        )
        tmp = tempfile.NamedTemporaryFile(suffix=".xml", delete=False)
        tmp.close()
        survey.print_xform_to_file(tmp.name, warnings=warnings)
        self.assertTrue(len(warnings) == 1)
        warning_expected = (
            "[row : 2] simserial is no longer supported on most devices. "
            "Only old versions of Collect on Android versions older than 11 still support it."
        )
        self.assertEqual(warning_expected, warnings[0])
        os.unlink(tmp.name)

    def test_subscriber_id_deprecation_warning(self):
        warnings = []
        survey = self.md_to_pyxform_survey(
            """
            | survey |              |                          |                                            |
            |        | type         | name                     | label                                      |
            |        | subscriberid | subscriberid             | sub id - extra warning generated w/o this  |
            |        | note         | subscriberid_test_output | subscriberid_test_output: ${subscriberid}  |
            """,
            warnings=warnings,
        )
        tmp = tempfile.NamedTemporaryFile(suffix=".xml", delete=False)
        tmp.close()
        survey.print_xform_to_file(tmp.name, warnings=warnings)
        self.assertTrue(len(warnings) == 1)
        warning_expected = (
            "[row : 2] subscriberid is no longer supported on most devices. "
            "Only old versions of Collect on Android versions older than 11 still support it."
        )
        self.assertEqual(warning_expected, warnings[0])
        os.unlink(tmp.name)



# -*- coding: utf-8 -*-
"""
XForm Survey element classes for different question types.
"""
import os.path
import re

from pyxform.errors import PyXFormError
from pyxform.question_type_dictionary import QUESTION_TYPE_DICT
from pyxform.survey_element import SurveyElement
from pyxform.utils import (
    basestring,
    node,
    unicode,
    default_is_dynamic,
    has_dynamic_label,
)


class Question(SurveyElement):
    def validate(self):
        SurveyElement.validate(self)

        # make sure that the type of this question exists in the
        # question type dictionary.
        if self.type not in QUESTION_TYPE_DICT:
            raise PyXFormError("Unknown question type '%s'." % self.type)

    def xml_instance(self, **kwargs):
        survey = self.get_root()
        attributes = {}
        attributes.update(self.get("instance", {}))
        for key, value in attributes.items():
            attributes[key] = survey.insert_xpaths(value, self)

        if self.get("default") and not default_is_dynamic(self.default, self.type):
            return node(self.name, unicode(self.get("default")), **attributes)
        return node(self.name, **attributes)

    def xml_control(self):
        if self.type == "calculate" or (
            ("calculate" in self.bind or self.trigger) and not (self.label or self.hint)
        ):
            nested_setvalues = self.get_root().get_setvalues_for_question_name(
                self.name
            )
            if nested_setvalues:
                for setvalue in nested_setvalues:
                    msg = (
                        "The question ${%s} is not user-visible so it can't be used as a calculation trigger for question ${%s}."
                        % (self.name, setvalue[0])
                    )
                    raise PyXFormError(msg)
            return None

        xml_node = self.build_xml()

        if xml_node:
            self.nest_setvalues(xml_node)

        return xml_node

    def nest_setvalues(self, xml_node):
        nested_setvalues = self.get_root().get_setvalues_for_question_name(self.name)

        if nested_setvalues:
            for setvalue in nested_setvalues:
                setvalue_attrs = {
                    "ref": self.get_root()
                    .insert_xpaths("${%s}" % setvalue[0], self.get_root())
                    .strip(),
                    "event": "xforms-value-changed",
                }
                if not (setvalue[1] == ""):
                    setvalue_attrs["value"] = self.get_root().insert_xpaths(
                        setvalue[1], self
                    )

                setvalue_node = node("setvalue", **setvalue_attrs)

                xml_node.appendChild(setvalue_node)

    def build_xml(self):
        return None


class InputQuestion(Question):
    """
    This control string is the same for: strings, integers, decimals,
    dates, geopoints, barcodes ...
    """

    def build_xml(self):
        control_dict = self.control
        label_and_hint = self.xml_label_and_hint()
        survey = self.get_root()
        # Resolve field references in attributes
        for key, value in control_dict.items():
            control_dict[key] = survey.insert_xpaths(value, self)
        control_dict["ref"] = self.get_xpath()

        result = node(**control_dict)
        if label_and_hint:
            for element in self.xml_label_and_hint():
                result.appendChild(element)

        # Input types are used for selects with external choices sheets.
        if self["query"]:
            choice_filter = self.get("choice_filter")
            query = "instance('" + self["query"] + "')/root/item"
            choice_filter = survey.insert_xpaths(choice_filter, self, True)
            if choice_filter:
                query += "[" + choice_filter + "]"
            result.setAttribute("query", query)
        return result


class TriggerQuestion(Question):
    def build_xml(self):
        control_dict = self.control
        survey = self.get_root()
        # Resolve field references in attributes
        for key, value in control_dict.items():
            control_dict[key] = survey.insert_xpaths(value, self)
        control_dict["ref"] = self.get_xpath()
        return node("trigger", *self.xml_label_and_hint(), **control_dict)


class UploadQuestion(Question):
    def _get_media_type(self):
        return self.control["mediatype"]

    def build_xml(self):
        control_dict = self.control
        survey = self.get_root()
        # Resolve field references in attributes
        for key, value in control_dict.items():
            control_dict[key] = survey.insert_xpaths(value, self)
        control_dict["ref"] = self.get_xpath()
        control_dict["mediatype"] = self._get_media_type()
        return node("upload", *self.xml_label_and_hint(), **control_dict)


class Option(SurveyElement):
    def xml_value(self):
        return node("value", self.name)

    def xml(self):
        item = node("item")
        self.xml_label()
        item.appendChild(self.xml_label())
        item.appendChild(self.xml_value())

        return item

    def validate(self):
        pass


class MultipleChoiceQuestion(Question):
    def __init__(self, **kwargs):
        kwargs_copy = kwargs.copy()
        # Notice that choices can be specified under choices or children.
        # I'm going to try to stick to just choices.
        # Aliases in the json format will make it more difficult
        # to use going forward.
        choices = list(kwargs_copy.pop("choices", [])) + list(
            kwargs_copy.pop("children", [])
        )
        Question.__init__(self, **kwargs_copy)
        for choice in choices:
            self.add_choice(**choice)

    def add_choice(self, **kwargs):
        option = Option(**kwargs)
        self.add_child(option)

    def validate(self):
        Question.validate(self)
        descendants = self.iter_descendants()
        next(descendants)  # iter_descendants includes self; we need to pop it

        for choice in descendants:
            choice.validate()

    def build_xml(self):
        assert self.bind["type"] in ["string", "odk:rank"]
        survey = self.get_root()
        control_dict = self.control.copy()
        # Resolve field references in attributes
        for key, value in control_dict.items():
            control_dict[key] = survey.insert_xpaths(value, self)
        control_dict["ref"] = self.get_xpath()

        result = node(**control_dict)
        for element in self.xml_label_and_hint():
            result.appendChild(element)

        choices = survey.get("choices")
        multi_language = False
        if choices is not None and len(choices) > 0:
            first_choices = next(iter(choices.values()))
            multi_language = isinstance(first_choices[0].get("label"), dict)

        # itemset are only supposed to be strings,
        # check to prevent the rare dicts that show up
        if self["itemset"] and isinstance(self["itemset"], basestring):
            choice_filter = self.get("choice_filter")
            itemset_value_ref = "name"
            itemset, file_extension = os.path.splitext(self["itemset"])
            has_media = False
            has_dyn_label = False
            is_previous_question = bool(re.match(r"^\${.*}$", self.get("itemset")))

            if choices.get(itemset):
                has_media = bool(choices[itemset][0].get("media"))
                has_dyn_label = has_dynamic_label(choices[itemset], multi_language)

            if file_extension in [".csv", ".xml"]:
                itemset = itemset
                itemset_label_ref = "label"
            else:
                if not multi_language and not has_media and not has_dyn_label:
                    itemset = self["itemset"]
                    itemset_label_ref = "label"
                else:
                    itemset = self["itemset"]
                    itemset_label_ref = "jr:itext(itextId)"

            choice_filter = survey.insert_xpaths(
                choice_filter, self, True, is_previous_question
            )
            if is_previous_question:
                path = (
                    survey.insert_xpaths(self["itemset"], self, reference_parent=True)
                    .strip()
                    .split("/")
                )
                nodeset = "/".join(path[:-1])
                itemset_value_ref = path[-1]
                itemset_label_ref = path[-1]
                if choice_filter:
                    choice_filter = choice_filter.replace(
                        "current()/" + nodeset, "."
                    ).replace(nodeset, ".")
                else:
                    # Choices must have a value. Filter out repeat instances without
                    # an answer for the linked question
                    name = path[-1]
                    choice_filter = f"./{name} != ''"
            else:
                nodeset = "instance('" + itemset + "')/root/item"

            if choice_filter:
                nodeset += "[" + choice_filter + "]"

            if self["parameters"]:
                params = self["parameters"]

                if "randomize" in params and params["randomize"] == "true":
                    nodeset = "randomize(" + nodeset

                    if "seed" in params:
                        if params["seed"].startswith("${"):
                            nodeset = (
                                nodeset
                                + ", "
                                + survey.insert_xpaths(params["seed"], self).strip()
                            )
                        else:
                            nodeset = nodeset + ", " + params["seed"]

                    nodeset += ")"

            itemset_children = [
                node("value", ref=itemset_value_ref),
                node("label", ref=itemset_label_ref),
            ]
            result.appendChild(node("itemset", *itemset_children, nodeset=nodeset))
        else:
            for child in self.children:
                result.appendChild(child.xml())

        return result


class SelectOneQuestion(MultipleChoiceQuestion):
    def __init__(self, **kwargs):
        super(SelectOneQuestion, self).__init__(**kwargs)
        self._dict[self.TYPE] = "select one"


class Tag(SurveyElement):
    def __init__(self, **kwargs):
        kwargs_copy = kwargs.copy()
        choices = kwargs_copy.pop("choices", []) + kwargs_copy.pop("children", [])

        super(Tag, self).__init__(**kwargs_copy)

        if choices:
            self.children = []

            for choice in choices:
                option = Option(**choice)
                self.add_child(option)

    def xml(self):
        result = node("tag", key=self.name)
        self.xml_label()
        result.appendChild(self.xml_label())
        for choice in self.children:
            result.appendChild(choice.xml())

        return result

    def validate(self):
        pass


class OsmUploadQuestion(UploadQuestion):
    def __init__(self, **kwargs):
        kwargs_copy = kwargs.copy()
        tags = kwargs_copy.pop("tags", []) + kwargs_copy.pop("children", [])

        super(OsmUploadQuestion, self).__init__(**kwargs_copy)

        if tags:
            self.children = []

            for tag in tags:
                self.add_tag(**tag)

    def add_tag(self, **kwargs):
        tag = Tag(**kwargs)
        self.add_child(tag)

    def build_xml(self):
        control_dict = self.control
        control_dict["ref"] = self.get_xpath()
        control_dict["mediatype"] = self._get_media_type()
        result = node("upload", *self.xml_label_and_hint(), **control_dict)

        for osm_tag in self.children:
            result.appendChild(osm_tag.xml())

        return result


class RangeQuestion(Question):
    def build_xml(self):
        control_dict = self.control
        label_and_hint = self.xml_label_and_hint()
        survey = self.get_root()
        # Resolve field references in attributes
        for key, value in control_dict.items():
            control_dict[key] = survey.insert_xpaths(value, self)
        control_dict["ref"] = self.get_xpath()
        params = self.get("parameters", {})
        control_dict.update(params)
        result = node(**control_dict)
        if label_and_hint:
            for element in self.xml_label_and_hint():
                result.appendChild(element)

        return result



# -*- coding: utf-8 -*-
"""
Test field-list labels
"""
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class FieldListLabels(PyxformTestCase):
    """Test unlabeled group"""

    def test_unlabeled_group(self):
        warnings = []

        survey = self.md_to_pyxform_survey(
            """
            | survey |             |          |         |
            |        | type        | name     | label   |
            |        | begin_group | my-group |         |
            |        | text        | my-text  | my-text |
            |        | end_group   |          |         |
            """,
            warnings=warnings,
        )

        self.assertTrue(len(warnings) == 1)
        self.assertTrue("[row : 2] Group has no label" in warnings[0])

    def test_unlabeled_group_alternate_syntax(self):
        warnings = []

        survey = self.md_to_pyxform_survey(
            """
            | survey |             |          |                     |
            |        | type        | name     | label::English (en) |
            |        | begin group | my-group |                     |
            |        | text        | my-text  | my-text             |
            |        | end group   |          |                     |
            """,
            warnings=warnings,
        )

        self.assertTrue(len(warnings) == 1)
        self.assertTrue("[row : 2] Group has no label" in warnings[0])

    def test_unlabeled_group_fieldlist(self):
        warnings = []

        survey = self.md_to_pyxform_survey(
            """
            | survey |              |           |         |            |
            |        | type         | name      | label   | appearance |
            |        | begin_group  | my-group  |         | field-list |
            |        | text         | my-text   | my-text |            |
            |        | end_group    |           |         |            |
            """,
            warnings=warnings,
        )

        self.assertTrue(len(warnings) == 0)

    def test_unlabeled_group_fieldlist_alternate_syntax(self):
        warnings = []

        survey = self.md_to_pyxform_survey(
            """
            | survey |              |           |         |            |
            |        | type         | name      | label   | appearance |
            |        | begin group  | my-group  |         | field-list |
            |        | text         | my-text   | my-text |            |
            |        | end group    |           |         |            |
            """,
            warnings=warnings,
        )

        self.assertTrue(len(warnings) == 0)

    def test_unlabeled_repeat(self):
        warnings = []

        survey = self.md_to_pyxform_survey(
            """
            | survey |              |           |         |
            |        | type         | name      | label   |
            |        | begin_repeat | my-repeat |         |
            |        | text         | my-text   | my-text |
            |        | end_repeat   |           |         |
            """,
            warnings=warnings,
        )

        self.assertTrue(len(warnings) == 1)
        self.assertTrue("[row : 2] Repeat has no label" in warnings[0])

    def test_unlabeled_repeat_fieldlist(self):
        warnings = []

        survey = self.md_to_pyxform_survey(
            """
            | survey |              |           |         |            |
            |        | type         | name      | label   | appearance |
            |        | begin_repeat | my-repeat |         | field-list |
            |        | text         | my-text   | my-text |            |
            |        | end_repeat   |           |         |            |
            """,
            warnings=warnings,
        )

        self.assertTrue(len(warnings) == 1)
        self.assertTrue("[row : 2] Repeat has no label" in warnings[0])



# -*- coding: utf-8 -*-
import os
import posixpath
import threading

try:
    from http.server import SimpleHTTPRequestHandler
    from socketserver import ThreadingTCPServer
    from urllib.parse import unquote
except ImportError:
    from SimpleHTTPServer import SimpleHTTPRequestHandler
    from SocketServer import ThreadingTCPServer
    from urllib2 import unquote


HERE = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")


class SimpleHTTPRequestHandlerHere(SimpleHTTPRequestHandler, object):
    def send_head(self):
        if self.client_address[0] != "127.0.0.1":
            self.send_error(
                401, "Unauthorized", "No permission -- see authorization schemes"
            )
            return None
        else:
            return super(SimpleHTTPRequestHandlerHere, self).send_head()

    def translate_path(self, path):
        """
        Ignore the actual request path and just serve a specific folder.

        Mostly same as py3.6 source, replacing "path = os.getcwd()" with HERE so
        that the directory list or file transfers are relative to HERE rather
        than the current working directory.
        """
        # abandon query parameters
        path = path.split("?", 1)[0]
        path = path.split("#", 1)[0]
        # Don"t forget explicit trailing slash when normalizing. Issue17324
        trailing_slash = path.rstrip().endswith("/")
        try:
            path = unquote(path, errors="surrogatepass")
        except UnicodeDecodeError:
            path = unquote(path)
        except TypeError:  # py2 only accepts one param.
            path = unquote(path)
        path = posixpath.normpath(path)
        words = path.split("/")
        words = filter(None, words)
        path = HERE  # edited
        for word in words:
            if os.path.dirname(word) or word in (os.curdir, os.pardir):
                # Ignore components that are not a simple file/directory name
                continue
            path = os.path.join(path, word)
        if trailing_slash:
            path += "/"
        return path

    def log_message(self, format, *args):
        pass  # be quiet


class ThreadingServerInThread(object):
    """
    Context manager for running a threading http server in a thread.

    Since the Thread is not using "daemon=True", it will keep Python running
    until the context manager exits, which means until request completion.
    """

    def __init__(self, port=8000):
        self._server_address = ("127.0.0.1", port)
        self._handler = SimpleHTTPRequestHandlerHere
        self.httpd = ThreadingTCPServer(
            self._server_address, self._handler, bind_and_activate=False
        )

    def _bind_and_activate(self):
        try:
            self.httpd.server_bind()
            self.httpd.server_activate()
        except Exception as e:
            self.httpd.server_close()
            raise e

    def start(self):
        self._bind_and_activate()
        thread = threading.Thread(target=self.httpd.serve_forever)
        thread.start()

    def stop(self):
        self.httpd.shutdown()
        self.httpd.server_close()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()



# -*- coding: utf-8 -*-
"""
pyxform utils module.
"""
import codecs
import copy
import json
import os
import re
from xml.dom.minidom import Element, Text, parseString

import unicodecsv as csv
import xlrd

try:
    from json.decoder import JSONDecodeError
except ImportError:
    # json module raises a ValueError exception when it encounters an invalid JSON
    JSONDecodeError = ValueError

try:
    unicode("str")
except NameError:
    unicode = str
    basestring = str
    unichr = chr
else:
    try:
        unicode = unicode
        basestring = basestring
        unichr = unichr
    except NameError:  # special cases where unicode is defined in python3
        unicode = str
        basestring = str
        unichr = chr

SEP = "_"

# http://www.w3.org/TR/REC-xml/
TAG_START_CHAR = r"[a-zA-Z:_]"
TAG_CHAR = r"[a-zA-Z:_0-9\-.]"
XFORM_TAG_REGEXP = "%(start)s%(char)s*" % {"start": TAG_START_CHAR, "char": TAG_CHAR}

INVALID_XFORM_TAG_REGEXP = r"[^a-zA-Z:_][^a-zA-Z:_0-9\-.]*"

LAST_SAVED_INSTANCE_NAME = "__last-saved"
BRACKETED_TAG_REGEX = re.compile(r"\${(last-saved#)?(.*?)}")
LAST_SAVED_REGEX = re.compile(r"\${last-saved#(.*?)}")

NSMAP = {
    "xmlns": "http://www.w3.org/2002/xforms",
    "xmlns:h": "http://www.w3.org/1999/xhtml",
    "xmlns:ev": "http://www.w3.org/2001/xml-events",
    "xmlns:xsd": "http://www.w3.org/2001/XMLSchema",
    "xmlns:jr": "http://openrosa.org/javarosa",
    "xmlns:orx": "http://openrosa.org/xforms",
    "xmlns:odk": "http://www.opendatakit.org/xforms",
}


class DetachableElement(Element):
    """
    Element classes are not meant to be instantiated directly.   This
    restriction was not strictly enforced in Python 2, but is effectively
    enforced in Python 3 via http://bugs.python.org/issue15290.

    A simple workaround (for now) is to set an extra attribute on the Element
    class.  The long term fix will probably be to rewrite node() to use
    document.createElement rather than Element.
    """

    def __init__(self, *args, **kwargs):
        Element.__init__(self, *args, **kwargs)
        self.ownerDocument = None


class PatchedText(Text):
    def writexml(self, writer, indent="", addindent="", newl=""):
        """Same as original but no replacing double quotes with '&quot;'."""
        data = "%s%s%s" % (indent, self.data, newl)
        if data:
            data = data.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")
        writer.write(data)


def is_valid_xml_tag(tag):
    """
    Use a regex to see if there are any invalid characters (i.e. spaces).
    """
    return re.search(r"^" + XFORM_TAG_REGEXP + r"$", tag)


def node(*args, **kwargs):
    """
    args[0] -- a XML tag
    args[1:] -- an array of children to append to the newly created node
            or if a unicode arg is supplied it will be used to make a text node
    kwargs -- attributes
    returns a xml.dom.minidom.Element
    """
    blocked_attributes = ["tag"]
    tag = args[0] if len(args) > 0 else kwargs["tag"]
    args = args[1:]
    result = DetachableElement(tag)
    unicode_args = [u for u in args if type(u) == unicode]
    assert len(unicode_args) <= 1
    parsed_string = False

    # Convert the kwargs xml attribute dictionary to a xml.dom.minidom.Element. Sort the
    # attributes to guarantee a consistent order across Python versions.
    # See pyxform_test_case.reorder_attributes for details.
    for k, v in iter(sorted(kwargs.items())):
        if k in blocked_attributes:
            continue
        if k == "toParseString":
            if v is True and len(unicode_args) == 1:
                parsed_string = True
                # Add this header string so parseString can be used?
                s = (
                    '<?xml version="1.0" ?><'
                    + tag
                    + ">"
                    + unicode_args[0]
                    + "</"
                    + tag
                    + ">"
                )
                parsed_node = parseString(s.encode("utf-8")).documentElement
                # Move node's children to the result Element
                # discarding node's root
                for child in parsed_node.childNodes:
                    result.appendChild(copy.deepcopy(child))
        else:
            result.setAttribute(k, v)

    if len(unicode_args) == 1 and not parsed_string:
        text_node = PatchedText()
        text_node.data = unicode_args[0]
        result.appendChild(text_node)
    for n in args:
        if type(n) == int or type(n) == float or type(n) == bytes:
            text_node = PatchedText()
            text_node.data = unicode(n)
            result.appendChild(text_node)
        elif type(n) is not unicode:
            result.appendChild(n)
    return result


def get_pyobj_from_json(str_or_path):
    """
    This function takes either a json string or a path to a json file,
    it loads the json into memory and returns the corresponding Python
    object.
    """
    try:
        # see if treating str_or_path as a path works
        fp = codecs.open(str_or_path, mode="r", encoding="utf-8")
        doc = json.load(fp, encoding="utf-8")
    except (IOError, JSONDecodeError, OSError):
        # if it doesn't work load the text
        doc = json.loads(str_or_path)
    return doc


def flatten(li):
    for subli in li:
        for it in subli:
            yield it


def sheet_to_csv(workbook_path, csv_path, sheet_name):
    from pyxform.xls2json_backends import xls_value_to_unicode

    wb = xlrd.open_workbook(workbook_path)
    try:
        sheet = wb.sheet_by_name(sheet_name)
    except xlrd.biffh.XLRDError:
        return False
    if not sheet or sheet.nrows < 2:
        return False
    with open(csv_path, "wb") as f:
        writer = csv.writer(f, quoting=csv.QUOTE_ALL)
        mask = [v and len(v.strip()) > 0 for v in sheet.row_values(0)]
        for row_idx in range(sheet.nrows):
            csv_data = []
            try:
                for v, m in zip(sheet.row(row_idx), mask):
                    if m:
                        value = v.value
                        value_type = v.ctype
                        data = xls_value_to_unicode(value, value_type, wb.datemode)
                        # clean the values of leading and trailing whitespaces
                        data = data.strip()
                        csv_data.append(data)
            except TypeError:
                continue
            writer.writerow(csv_data)

    return True


def has_external_choices(json_struct):
    """
    Returns true if a select one external prompt is used in the survey.
    """
    if isinstance(json_struct, dict):
        for k, v in json_struct.items():
            if (
                k == "type"
                and isinstance(v, basestring)
                and v.startswith("select one external")
            ):
                return True
            elif has_external_choices(v):
                return True
    elif isinstance(json_struct, list):
        for v in json_struct:
            if has_external_choices(v):
                return True
    return False


def get_languages_with_bad_tags(languages):
    """
    Returns languages with invalid or missing IANA subtags.
    """
    iana_subtags = []
    with open(os.path.join(os.path.dirname(__file__), "iana_subtags.txt")) as f:
        iana_subtags = f.read().splitlines()

    lang_code_regex = re.compile(r"\((.*)\)$")

    languages_with_bad_tags = []
    for lang in languages:
        lang_code = re.search(lang_code_regex, lang)

        if lang != "default" and (
            not (lang_code) or not (lang_code.group(1) in iana_subtags)
        ):
            languages_with_bad_tags.append(lang)
    return languages_with_bad_tags


def default_is_dynamic(element_default, element_type=None):
    """
    Returns true if the default value is a dynamic value.

    Dynamic value for now is defined as:
    * Contains arithmetic operator, including 'div' and 'mod' (except '-' for 'date' type).
    * Contains brackets, parentheses or braces.
    """
    if not isinstance(element_default, basestring):
        return False

    dynamic_markers = {
        " mod ",
        " div ",
        "*",
        "|",
        "+",
        "-",
        "[",
        "]",
        "{",
        "}",
        "(",
        ")",
    }
    if element_type is not None and element_type == "date":
        dynamic_markers.remove("-")

    return any(s in element_default for s in dynamic_markers)


# If the first or second choice label includes a reference, we must use itext.
# Check the first two choices in case first is something like "Other".
def has_dynamic_label(choice_list, multi_language):
    if not multi_language:
        for i in range(0, min(2, len(choice_list))):
            if (
                choice_list[i].get("label") is not None
                and re.search(BRACKETED_TAG_REGEX, choice_list[i].get("label"))
                is not None
            ):
                return True
    return False


def levenshtein_distance(a: str, b: str) -> int:
    """
    Calculate Levenshtein distance between two strings.

    A Python translation of the "iterative with two matrix rows" algorithm from
    Wikipedia: https://en.wikipedia.org/wiki/Levenshtein_distance

    :param a: The first string to compare.
    :param b: The second string to compare.
    :return:
    """
    m = len(a)
    n = len(b)

    # create two work vectors of integer distances
    v1 = [0 for i in range(0, n + 1)]

    # initialize v0 (the previous row of distances)
    # this row is A[0][i]: edit distance for an empty s
    # the distance is just the number of characters to delete from t
    v0 = [i for i in range(0, n + 1)]

    for i in range(0, m):
        # calculate v1 (current row distances) from the previous row v0

        # first element of v1 is A[i+1][0]
        #   edit distance is delete (i+1) chars from s to match empty t
        v1[0] = i + 1

        # use formula to fill in the rest of the row
        for j in range(0, n):
            # calculating costs for A[i+1][j+1]
            deletion_cost = v0[j + 1] + 1
            insertion_cost = v1[j] + 1
            if a[i] == b[j]:
                substitution_cost = v0[j]
            else:
                substitution_cost = v0[j] + 1

            v1[j + 1] = min((deletion_cost, insertion_cost, substitution_cost))

        # copy v1 (current row) to v0 (previous row) for next iteration
        # since data in v1 is always invalidated, a swap without copy could be more efficient
        v0 = copy.copy(v1)
    # after the last swap, the results of v1 are now in v0
    return v0[n]



# -*- coding: utf-8 -*-
"""
odk_validate.py
A python wrapper around ODK Validate
"""
from __future__ import print_function, unicode_literals

import logging
import os
import sys
import shutil
from typing import TYPE_CHECKING
from pyxform.validators.error_cleaner import ErrorCleaner
from pyxform.validators.util import (
    XFORM_SPEC_PATH,
    check_readable,
    run_popen_with_timeout,
)


if TYPE_CHECKING:
    from pyxform.validators.util import PopenResult


CURRENT_DIRECTORY = os.path.dirname(os.path.realpath(__file__))
ODK_VALIDATE_PATH = os.path.join(CURRENT_DIRECTORY, "bin", "ODK_Validate.jar")


class ODKValidateError(Exception):
    """ODK Validation exception error."""

    pass


def install_exists():
    """Returns True if ODK_VALIDATE_PATH exists."""
    return os.path.exists(ODK_VALIDATE_PATH)


def _call_validator(path_to_xform, bin_file_path=ODK_VALIDATE_PATH) -> "PopenResult":
    return run_popen_with_timeout(
        ["java", "-Djava.awt.headless=true", "-jar", bin_file_path, path_to_xform], 100
    )


def install_ok(bin_file_path=ODK_VALIDATE_PATH):
    """
    Check if ODK Validate functions as expected.
    """
    check_readable(file_path=XFORM_SPEC_PATH)
    result = _call_validator(
        path_to_xform=XFORM_SPEC_PATH, bin_file_path=bin_file_path,
    )
    if result.return_code == 1:
        return False

    return True


def check_java_available():
    """
    Check if 'which java' returncode is 0. If not, raise an error since java is required.
    """
    java_path = shutil.which(cmd="java")
    if java_path is not None:
        return
    msg = (
        "Form validation failed because Java (8+ required) could not be found. "
        "To fix this, please either: 1) install Java, or 2) run pyxform with the "
        "--skip_validate flag, or 3) add the installed Java to the environment path."
    )
    raise EnvironmentError(msg)


def check_xform(path_to_xform):
    """Run ODK Validate against the XForm in `path_to_xform`.

    Returns an array of warnings if the form is valid.
    Throws an exception if it is not
    Does not do a LBYL check for compatible java version as per pyxform/#481
    """
    # check for available java version
    check_java_available()

    # resultcode indicates validity of the form
    # timeout indicates whether validation ran out of time to complete
    # stdout is not used because it has some warnings that always
    # appear and can be ignored.
    # stderr is treated as a warning if the form is valid or an error
    # if it is invalid.
    result = _call_validator(path_to_xform=path_to_xform)
    warnings = []

    if result.timeout:
        return ["XForm took to long to completely validate."]
    else:
        if result.return_code > 0:  # Error invalid
            raise ODKValidateError(
                "ODK Validate Errors:\n" + ErrorCleaner.odk_validate(result.stderr)
            )
        elif result.return_code == 0:
            if result.stderr:
                warnings.append("ODK Validate Warnings:\n" + result.stderr)
        elif result.return_code < 0:
            return ["Bad return code from ODK Validate."]

    return warnings


if __name__ == "__main__":
    logger = logging.getLogger(__name__)
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(logging.INFO)
    logger.info(__doc__)

    check_xform(sys.argv[1])



# -*- coding: utf-8 -*-
"""
Survey Element base class for all survey elements.
"""
import json
import re

from pyxform import constants
from pyxform.errors import PyXFormError
from pyxform.question_type_dictionary import QUESTION_TYPE_DICT
from pyxform.utils import (
    INVALID_XFORM_TAG_REGEXP,
    BRACKETED_TAG_REGEX,
    is_valid_xml_tag,
    node,
    unicode,
    default_is_dynamic,
)
from pyxform.xls2json import print_pyobj_to_json
from typing import TYPE_CHECKING

try:
    from functools import lru_cache
except ImportError:
    from functools32 import lru_cache


if TYPE_CHECKING:
    from pyxform.utils import DetachableElement
    from typing import List


def _overlay(over, under):
    if type(under) == dict:
        result = under.copy()
        result.update(over)
        return result
    return over if over else under


class SurveyElement(dict):
    """
    SurveyElement is the base class we'll looks for the following keys
    in kwargs: name, label, hint, type, bind, control, parent,
    children, and question_type_dictionary.
    """

    # the following are important keys for the underlying dict that
    # describes this survey element
    FIELDS = {
        "name": unicode,
        constants.COMPACT_TAG: unicode,  # used for compact (sms) representation
        "sms_field": unicode,
        "sms_option": unicode,
        "label": unicode,
        "hint": unicode,
        "guidance_hint": unicode,
        "default": unicode,
        "type": unicode,
        "appearance": unicode,
        "parameters": unicode,
        "intent": unicode,
        "jr:count": unicode,
        "bind": dict,
        "instance": dict,
        "control": dict,
        "media": dict,
        # this node will also have a parent and children, like a tree!
        "parent": lambda: None,
        "children": list,
        "itemset": unicode,
        "choice_filter": unicode,
        "query": unicode,
        "autoplay": unicode,
        "flat": lambda: False,
        "action": unicode,
        "list_name": unicode,
        "trigger": unicode,
    }

    def _default(self):
        # TODO: need way to override question type dictionary
        defaults = QUESTION_TYPE_DICT
        return defaults.get(self.get("type"), {})

    def __getattr__(self, key):
        """
        Get attributes from FIELDS rather than the class.
        """
        if key in self.FIELDS:
            question_type_dict = self._default()
            under = question_type_dict.get(key, None)
            over = self.get(key)
            if not under:
                return over
            return _overlay(over, under)
        raise AttributeError(key)

    def __hash__(self):
        return hash(id(self))

    @property
    def __name__(self):
        return "SurveyElement"

    def __setattr__(self, key, value):
        self[key] = value

    def __init__(self, **kwargs):
        for key, default in self.FIELDS.items():
            self[key] = kwargs.get(key, default())
        self._link_children()

        # Create a space label for unlabeled elements with the label
        # appearance tag. # This is because such elements are used to label the
        # options for selects in a field-list and might want blank labels for
        # themselves.
        if self.control.get("appearance") == "label" and not self.label:
            self["label"] = " "

    def _link_children(self):
        for child in self.children:
            child.parent = self

    def add_child(self, child):
        self.children.append(child)
        child.parent = self

    def add_children(self, children):
        if type(children) == list:
            for child in children:
                self.add_child(child)
        else:
            self.add_child(children)

    binding_conversions = {
        "yes": "true()",
        "Yes": "true()",
        "YES": "true()",
        "true": "true()",
        "True": "true()",
        "TRUE": "true()",
        "no": "false()",
        "No": "false()",
        "NO": "false()",
        "false": "false()",
        "False": "false()",
        "FALSE": "false()",
    }

    CONVERTIBLE_BIND_ATTRIBUTES = [
        "readonly",
        "required",
        "relevant",
        "constraint",
        "calculate",
    ]

    # Supported media types for attaching to questions
    SUPPORTED_MEDIA = ["image", "audio", "video"]

    def validate(self):
        if not is_valid_xml_tag(self.name):
            invalid_char = re.search(INVALID_XFORM_TAG_REGEXP, self.name)
            msg = (
                "The name '{}' is an invalid XML tag, it contains an "
                "invalid character '{}'. Names must begin with a letter, "
                "colon, or underscore, subsequent characters can include "
                "numbers, dashes, and periods".format(self.name, invalid_char.group(0))
            )
            raise PyXFormError(msg)

    # TODO: Make sure renaming this doesn't cause any problems
    def iter_descendants(self):
        """
        A survey_element is a dictionary of survey_elements
        This method does a preorder traversal over them.
        For the time being this survery_element is included among its
        descendants
        """
        # it really seems like this method should not yield self
        yield self
        for e in self.children:
            for f in e.iter_descendants():
                yield f

    @lru_cache(maxsize=None)
    def any_repeat(self, parent_xpath):
        """Return True if there ia any repeat in `parent_xpath`."""
        for item in self.iter_descendants():
            if item.get_xpath() == parent_xpath and item.type == constants.REPEAT:
                return True

        return False

    def get_lineage(self):
        """
        Return a the list [root, ..., self._parent, self]
        """
        result = [self]
        current_element = self
        while current_element.parent:
            current_element = current_element.parent
            result = [current_element] + result
        # For some reason the root element has a True flat property...
        output = [result[0]]
        for item in result[1:]:
            if not item.get("flat"):
                output.append(item)
        return output

    def get_root(self):
        return self.get_lineage()[0]

    def get_xpath(self):
        """
        Return the xpath of this survey element.
        """
        return "/".join([""] + [n.name for n in self.get_lineage()])

    def get_abbreviated_xpath(self):
        lineage = self.get_lineage()
        if len(lineage) >= 2:
            return "/".join([unicode(n.name) for n in lineage[1:]])
        else:
            return lineage[0].name

    def _delete_keys_from_dict(self, dictionary: dict, keys: list):
        """
        Deletes a list of keys from a dictionary.
        Credits: https://stackoverflow.com/a/49723101
        """
        for key in keys:
            if key in dictionary:
                del dictionary[key]

        for value in dictionary.values():
            if isinstance(value, dict):
                self._delete_keys_from_dict(value, keys)

    def to_json_dict(self):
        """
        Create a dict copy of this survey element by removing inappropriate
        attributes and converting its children to dicts
        """
        self.validate()
        result = self.copy()
        to_delete = ["parent", "question_type_dictionary", "_created"]
        # Delete all keys that may cause a "Circular Reference"
        # error while converting the result to JSON
        self._delete_keys_from_dict(result, to_delete)
        children = result.pop("children")
        result["children"] = []
        for child in children:
            result["children"].append(child.to_json_dict())
        # remove any keys with empty values
        for k, v in list(result.items()):
            if not v:
                del result[k]

        return result

    def to_json(self):
        return json.dumps(self.to_json_dict())

    def json_dump(self, path=""):
        if not path:
            path = self.name + ".json"
        print_pyobj_to_json(self.to_json_dict(), path)

    def __eq__(self, y):
        return (
            hasattr(y, "to_json_dict")
            and callable(y.to_json_dict)
            and self.to_json_dict() == y.to_json_dict()
        )

    def _translation_path(self, display_element):
        return self.get_xpath() + ":" + display_element

    def get_translations(self, default_language):
        """
        Returns translations used by this element so they can be included in
        the <itext> block. @see survey._setup_translations
        """
        bind_dict = self.get("bind")
        if bind_dict and type(bind_dict) is dict:
            constraint_msg = bind_dict.get("jr:constraintMsg")
            if type(constraint_msg) is dict:
                for lang, text in constraint_msg.items():
                    yield {
                        "path": self._translation_path("jr:constraintMsg"),
                        "lang": lang,
                        "text": text,
                        "output_context": self,
                    }
            elif constraint_msg and re.search(BRACKETED_TAG_REGEX, constraint_msg):
                yield {
                    "path": self._translation_path("jr:constraintMsg"),
                    "lang": default_language,
                    "text": constraint_msg,
                    "output_context": self,
                }

            required_msg = bind_dict.get("jr:requiredMsg")
            if type(required_msg) is dict:
                for lang, text in required_msg.items():
                    yield {
                        "path": self._translation_path("jr:requiredMsg"),
                        "lang": lang,
                        "text": text,
                        "output_context": self,
                    }
            elif required_msg and re.search(BRACKETED_TAG_REGEX, required_msg):
                yield {
                    "path": self._translation_path("jr:requiredMsg"),
                    "lang": default_language,
                    "text": required_msg,
                    "output_context": self,
                }
            no_app_error_string = bind_dict.get("jr:noAppErrorString")
            if type(no_app_error_string) is dict:
                for lang, text in no_app_error_string.items():
                    yield {
                        "path": self._translation_path("jr:noAppErrorString"),
                        "lang": lang,
                        "text": text,
                        "output_context": self,
                    }

        for display_element in ["label", "hint", "guidance_hint"]:
            label_or_hint = self[display_element]

            if (
                display_element == "label"
                and self.needs_itext_ref()
                and type(label_or_hint) is not dict
                and label_or_hint
            ):
                label_or_hint = {default_language: label_or_hint}

            # always use itext for guidance hints because that's
            # how they're defined - https://opendatakit.github.io/xforms-spec/#languages
            if (
                display_element == "guidance_hint"
                and not (isinstance(label_or_hint, dict))
                and len(label_or_hint) > 0
            ):
                label_or_hint = {default_language: label_or_hint}

            # always use itext for hint if there's a guidance hint
            if (
                display_element == "hint"
                and not (isinstance(label_or_hint, dict))
                and len(label_or_hint) > 0
                and "guidance_hint" in self.keys()
                and len(self["guidance_hint"]) > 0
            ):
                label_or_hint = {default_language: label_or_hint}

            if type(label_or_hint) is dict:
                for lang, text in label_or_hint.items():
                    yield {
                        "display_element": display_element,  # Not used
                        "path": self._translation_path(display_element),
                        "element": self,  # Not used
                        "output_context": self,
                        "lang": lang,
                        "text": text,
                    }

    def get_media_keys(self):
        """
        @deprected
        I'm leaving this in just in case it has outside references.
        """
        return {"media": "%s:media" % self.get_xpath()}

    def needs_itext_ref(self):
        return type(self.label) is dict or (
            type(self.media) is dict and len(self.media) > 0
        )

    def get_setvalue_node_for_dynamic_default(self, in_repeat=False):
        if not self.default or not default_is_dynamic(self.default, self.type):
            return None

        default_with_xpath_paths = self.get_root().insert_xpaths(self.default, self)

        triggering_events = "odk-instance-first-load"
        if in_repeat:
            triggering_events = triggering_events + " odk-new-repeat"

        return node(
            "setvalue",
            ref=self.get_xpath(),
            value=default_with_xpath_paths,
            event=triggering_events,
        )

    # XML generating functions, these probably need to be moved around.
    def xml_label(self):
        if self.needs_itext_ref():
            # If there is a dictionary label, or non-empty media dict,
            # then we need to make a label with an itext ref
            ref = "jr:itext('%s')" % self._translation_path("label")
            return node("label", ref=ref)
        else:
            survey = self.get_root()
            label, output_inserted = survey.insert_output_values(self.label, self)
            return node("label", label, toParseString=output_inserted)

    def xml_hint(self):
        if isinstance(self.hint, dict) or self.guidance_hint:
            path = self._translation_path("hint")
            return node("hint", ref="jr:itext('%s')" % path)
        else:
            hint, output_inserted = self.get_root().insert_output_values(
                self.hint, self
            )
            return node("hint", hint, toParseString=output_inserted)

    def xml_label_and_hint(self) -> "List[DetachableElement]":
        """
        Return a list containing one node for the label and if there
        is a hint one node for the hint.
        """
        result = []
        label_appended = False
        if self.label or self.media:
            result.append(self.xml_label())
            label_appended = True

        if self.hint or self.guidance_hint:
            if not label_appended:
                result.append(self.xml_label())
            result.append(self.xml_hint())

        msg = "The survey element named '%s' " "has no label or hint." % self.name
        if len(result) == 0:
            raise PyXFormError(msg)
        # Guidance hint alone is not OK since they may be hidden by default.
        if not any((self.label, self.media, self.hint)) and self.guidance_hint:
            raise PyXFormError(msg)

        return result

    def xml_binding(self):
        """
        Return the binding for this survey element.
        """
        survey = self.get_root()
        bind_dict = self.bind.copy()
        if self.get("flat"):
            # Don't generate bind element for flat groups.
            return None
        if bind_dict:
            # the expression goes in a setvalue action
            if self.trigger and "calculate" in self.bind:
                del bind_dict["calculate"]

            for k, v in bind_dict.items():
                # I think all the binding conversions should be happening on
                # the xls2json side.
                if (
                    hashable(v)
                    and v in self.binding_conversions
                    and k in self.CONVERTIBLE_BIND_ATTRIBUTES
                ):
                    v = self.binding_conversions[v]
                if k == "jr:constraintMsg" and (
                    type(v) is dict or re.search(BRACKETED_TAG_REGEX, v)
                ):
                    v = "jr:itext('%s')" % self._translation_path("jr:constraintMsg")
                if k == "jr:requiredMsg" and (
                    type(v) is dict or re.search(BRACKETED_TAG_REGEX, v)
                ):
                    v = "jr:itext('%s')" % self._translation_path("jr:requiredMsg")
                if k == "jr:noAppErrorString" and type(v) is dict:
                    v = "jr:itext('%s')" % self._translation_path("jr:noAppErrorString")
                bind_dict[k] = survey.insert_xpaths(v, context=self)
            return node("bind", nodeset=self.get_xpath(), **bind_dict)
        return None

    def xml_bindings(self):
        """
        Return a list of bindings for this node and all its descendants.
        """
        result = []
        for e in self.iter_descendants():
            xml_binding = e.xml_binding()
            if xml_binding is not None:
                result.append(xml_binding)

            # dynamic defaults for repeats go in the body. All other dynamic defaults (setvalue actions) go in the model
            if (
                len(
                    [
                        ancestor
                        for ancestor in e.get_lineage()
                        if ancestor.type == "repeat"
                    ]
                )
                == 0
            ):
                dynamic_default = e.get_setvalue_node_for_dynamic_default()
                if dynamic_default:
                    result.append(dynamic_default)
        return result

    def xml_control(self):
        """
        The control depends on what type of question we're asking, it
        doesn't make sense to implement here in the base class.
        """
        raise NotImplementedError("Control not implemented")

    def xml_action(self):
        """
        Return the action for this survey element.
        """
        if self.action:
            action_dict = self.action.copy()
            if action_dict:
                name = action_dict["name"]
                del action_dict["name"]
                return node(name, ref=self.get_xpath(), **action_dict)

        return None

    def xml_actions(self):
        """
        Return a list of actions for this node and all its descendants.
        """
        result = []
        for e in self.iter_descendants():
            xml_action = e.xml_action()
            if xml_action is not None:
                result.append(xml_action)
        return result


def hashable(v):
    """Determine whether `v` can be hashed."""
    try:
        hash(v)
    except TypeError:
        return False
    return True



import unittest
from pyxform.utils import levenshtein_distance


""" /* To test in Postgres */
CREATE EXTENSION IF NOT EXISTS fuzzystrmatch;
SELECT levenshtein(a, b)
FROM (
  VALUES
    ('sitting', 'kitten'),
    ('Sunday', 'Saturday'),
    ('settings', 'settings'),
    ('setting', 'settings'),
    ('abcdefghijklm', 'nopqrstuvwxyz'),
    ('abc  klm', '** _rs /wxyz'),
    ('ABCD', 'abcd')
) as t(a, b);
"""


class TestLevenshteinDistance(unittest.TestCase):
    def test_levenshtein_distance(self):
        """Should return the expected distance value."""
        # Verified against Postgres v10 extension "fuzzystrmatch" levenshtein().
        test_data = (
            (3, "sitting", "kitten"),
            (3, "Sunday", "Saturday"),
            (0, "settings", "settings"),
            (1, "setting", "settings"),
            (13, "abcdefghijklm", "nopqrstuvwxyz"),
            (11, "abc  klm", "** _rs /wxyz"),
            (4, "ABCD", "abcd"),
        )
        for i in test_data:
            self.assertEqual(i[0], levenshtein_distance(i[1], i[2]), str(i))



# -*- coding: utf-8 -*-
"""
xls2xform converts properly formatted Excel documents into XForms for
use with ODK Collect.
"""
from __future__ import print_function

import argparse
import json
import logging
import os
from os.path import splitext

from pyxform import builder, xls2json
from pyxform.utils import has_external_choices, sheet_to_csv
from pyxform.validators.odk_validate import ODKValidateError

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.INFO)


def get_xml_path(path):
    """
    Returns the xform file path

    Generates an output path for the xform file from the given
    xlsx input file path.
    """
    return splitext(path)[0] + ".xml"


def xls2xform_convert(
    xlsform_path, xform_path, validate=True, pretty_print=True, enketo=False
):
    warnings = []

    json_survey = xls2json.parse_file_to_json(xlsform_path, warnings=warnings)
    survey = builder.create_survey_element_from_dict(json_survey)
    # Setting validate to false will cause the form not to be processed by
    # ODK Validate.
    # This may be desirable since ODK Validate requires launching a subprocess
    # that runs some java code.
    survey.print_xform_to_file(
        xform_path,
        validate=validate,
        pretty_print=pretty_print,
        warnings=warnings,
        enketo=enketo,
    )
    output_dir = os.path.split(xform_path)[0]
    if has_external_choices(json_survey):
        itemsets_csv = os.path.join(output_dir, "itemsets.csv")
        choices_exported = sheet_to_csv(xlsform_path, itemsets_csv, "external_choices")
        if not choices_exported:
            warnings.append(
                "Could not export itemsets.csv, perhaps the "
                "external choices sheet is missing."
            )
        else:
            logger.info("External choices csv is located at: %s", itemsets_csv)
    return warnings


def _create_parser():
    """
    Parse command line arguments.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "path_to_XLSForm",
        help="Path to the Excel XLSX file with the XLSForm definition.",
    )
    parser.add_argument("output_path", help="Path to save the output to.", nargs="?")
    parser.add_argument(
        "--json",
        action="store_true",
        help="Capture everything and report in JSON format.",
    )
    parser.add_argument(
        "--skip_validate",
        action="store_false",
        default=True,
        help="Do not run any external validators on the output XForm XML. "
        "Without this flag, ODK Validate is run by default for backwards "
        "compatibility. Internal pyxform checks cannot be skipped.",
    )
    parser.add_argument(
        "--odk_validate",
        action="store_true",
        default=False,
        help="Run the ODK Validate XForm external validator.",
    )
    parser.add_argument(
        "--enketo_validate",
        action="store_true",
        default=False,
        help="Run the Enketo Validate XForm external validator.",
    )
    parser.add_argument(
        "--pretty_print",
        action="store_true",
        default=False,
        help="Print XML forms with collapsed whitespace instead of pretty-printed.",
    )
    return parser


def _validator_args_logic(args):
    """
    Implements logic for how validator arguments work in combination.

    As per: https://github.com/XLSForm/pyxform/pull/167#issuecomment-353382008

    **backwards-compatible**
    `xls2xform.py myform --skip_validate`: no validators
    `xls2xform.py myform`: ODK only

    **new**
    `xls2xform.py myform --enketo_validate`: Enketo only
    `xls2xform.py myform --odk_validate`: ODK only
    `xls2xform.py myform --enketo_validate --odk_validate`: both
    `xls2xform.py myform --enketo_validate --odk_validate --skip_validate`: no validators
    """
    if not args.skip_validate:
        args.odk_validate = False
        args.enketo_validate = False
    elif args.skip_validate and not (args.odk_validate or args.enketo_validate):
        args.odk_validate = True
        args.enketo_validate = False
    return args


def main_cli():
    parser = _create_parser()
    raw_args = parser.parse_args()
    args = _validator_args_logic(args=raw_args)

    # auto generate an output path if one was not given
    if args.output_path is None:
        args.output_path = get_xml_path(args.path_to_XLSForm)

    if args.json:
        # Store everything in a list just in case the user wants to output
        # as a JSON encoded string.
        response = {"code": None, "message": None, "warnings": []}

        try:
            response["warnings"] = xls2xform_convert(
                xlsform_path=args.path_to_XLSForm,
                xform_path=args.output_path,
                validate=args.odk_validate,
                pretty_print=args.pretty_print,
                enketo=args.enketo_validate,
            )

            response["code"] = 100
            response["message"] = "Ok!"

            if response["warnings"]:
                response["code"] = 101
                response["message"] = "Ok with warnings."

        except Exception as e:
            # Catch the exception by default.
            response["code"] = 999
            response["message"] = str(e)

        logger.info(json.dumps(response))
    else:
        try:
            warnings = xls2xform_convert(
                xlsform_path=args.path_to_XLSForm,
                xform_path=args.output_path,
                validate=args.odk_validate,
                pretty_print=args.pretty_print,
                enketo=args.enketo_validate,
            )
        except EnvironmentError as e:
            # Do not crash if 'java' not installed
            logger.error(e)
        except ODKValidateError as e:
            # Remove output file if there is an error
            os.remove(args.output_path)
            logger.error(e)
        else:
            if len(warnings) > 0:
                logger.warning("Warnings:")
            for w in warnings:
                logger.warning(w)
            logger.info("Conversion complete!")


if __name__ == "__main__":
    main_cli()



# -*- coding: utf-8 -*-
"""
Test builder module functionality.
"""
import os
import re
import xml.etree.ElementTree as ETree
from unittest import TestCase

from pyxform import InputQuestion, Survey
from pyxform.builder import SurveyElementBuilder, create_survey_from_xls
from pyxform.errors import PyXFormError
from pyxform.tests import utils
from pyxform.xls2json import print_pyobj_to_json

FIXTURE_FILETYPE = "xls"


class BuilderTests(TestCase):
    maxDiff = None

    #   Moving to spec tests
    #    def test_new_widgets(self):
    #        survey = utils.build_survey('widgets.xls')
    #        path = utils.path_to_text_fixture('widgets.xml')
    #        survey.to_xml
    #        with open(path) as f:
    #            expected = ETree.fromstring(survey.to_xml())
    #            result = ETree.fromstring(f.read())
    #            self.assertTrue(xml_compare(expected, result))

    def test_unknown_question_type(self):
        survey = utils.build_survey("unknown_question_type.xls")
        self.assertRaises(PyXFormError, survey.to_xml)

    def test_uniqueness_of_section_names(self):
        # Looking at the xls file, I think this test might be broken.
        survey = utils.build_survey("group_names_must_be_unique.xls")
        self.assertRaises(Exception, survey.to_xml)

    def setUp(self):
        self.this_directory = os.path.dirname(__file__)
        survey_out = Survey(name="age", sms_keyword="age", type="survey")
        question = InputQuestion(name="age")
        question.type = "integer"
        question.label = "How old are you?"
        survey_out.add_child(question)
        self.survey_out_dict = survey_out.to_json_dict()
        print_pyobj_to_json(
            self.survey_out_dict, utils.path_to_text_fixture("how_old_are_you.json")
        )

    def test_create_from_file_object(self):
        path = utils.path_to_text_fixture("yes_or_no_question.xls")
        with open(path, "rb") as f:
            create_survey_from_xls(f)

    def tearDown(self):
        fixture_path = utils.path_to_text_fixture("how_old_are_you.json")
        if os.path.exists(fixture_path):
            os.remove(fixture_path)

    def test_create_table_from_dict(self):
        d = {
            "type": "loop",
            "name": "my_loop",
            "label": {"English": "My Loop"},
            "columns": [
                {"name": "col1", "label": {"English": "column 1"}},
                {"name": "col2", "label": {"English": "column 2"}},
            ],
            "children": [
                {
                    "type": "integer",
                    "name": "count",
                    "label": {"English": "How many are there in this group?"},
                }
            ],
        }
        builder = SurveyElementBuilder()
        g = builder.create_survey_element_from_dict(d)

        expected_dict = {
            "name": "my_loop",
            "label": {"English": "My Loop"},
            "type": "group",
            "children": [
                {
                    "name": "col1",
                    "label": {"English": "column 1"},
                    "type": "group",
                    "children": [
                        {
                            "name": "count",
                            "label": {"English": "How many are there in this group?"},
                            "type": "integer",
                        }
                    ],
                },
                {
                    "name": "col2",
                    "label": {"English": "column 2"},
                    "type": "group",
                    "children": [
                        {
                            "name": "count",
                            "label": {"English": "How many are there in this group?"},
                            "type": "integer",
                        }
                    ],
                },
            ],
        }

        self.assertEqual(g.to_json_dict(), expected_dict)

    def test_specify_other(self):
        survey = utils.create_survey_from_fixture(
            "specify_other", filetype=FIXTURE_FILETYPE
        )
        expected_dict = {
            "name": "specify_other",
            "type": "survey",
            "title": "specify_other",
            "default_language": "default",
            "id_string": "specify_other",
            "sms_keyword": "specify_other",
            "children": [
                {
                    "name": "sex",
                    "label": {"English": "What sex are you?"},
                    "type": "select one",
                    "list_name": "sexes",
                    "children": [
                        # TODO Change to choices (there is stuff in the
                        # json2xform half that will need to change)
                        {"name": "male", "label": {"English": "Male"}},
                        {"name": "female", "label": {"English": "Female"}},
                        {"name": "other", "label": "Other"},
                    ],
                },
                {
                    "name": "sex_other",
                    "bind": {"relevant": "selected(../sex, 'other')"},
                    "label": "Specify other.",
                    "type": "text",
                },
                {
                    "children": [
                        {
                            "bind": {"jr:preload": "uid", "readonly": "true()"},
                            "name": "instanceID",
                            "type": "calculate",
                        }
                    ],
                    "control": {"bodyless": True},
                    "name": "meta",
                    "type": "group",
                },
            ],
        }
        self.maxDiff = None
        self.assertEqual(survey.to_json_dict(), expected_dict)

    def test_select_one_question_with_identical_choice_name(self):
        """
        testing to make sure that select ones whose choice names are the same
        as the name of the select one get compiled.
        """
        survey = utils.create_survey_from_fixture(
            "choice_name_same_as_select_name", filetype=FIXTURE_FILETYPE
        )
        expected_dict = {
            "name": "choice_name_same_as_select_name",
            "title": "choice_name_same_as_select_name",
            "sms_keyword": "choice_name_same_as_select_name",
            "default_language": "default",
            "id_string": "choice_name_same_as_select_name",
            "type": "survey",
            "children": [
                {
                    "children": [{"name": "zone", "label": "Zone"}],
                    "type": "select one",
                    "name": "zone",
                    "label": "Zone",
                    "list_name": "zone",
                },
                {
                    "children": [
                        {
                            "bind": {"jr:preload": "uid", "readonly": "true()"},
                            "name": "instanceID",
                            "type": "calculate",
                        }
                    ],
                    "control": {"bodyless": True},
                    "name": "meta",
                    "type": "group",
                },
            ],
        }
        self.maxDiff = None
        self.assertEqual(survey.to_json_dict(), expected_dict)

    def test_loop(self):
        survey = utils.create_survey_from_fixture("loop", filetype=FIXTURE_FILETYPE)
        expected_dict = {
            "name": "loop",
            "id_string": "loop",
            "sms_keyword": "loop",
            "title": "loop",
            "type": "survey",
            "default_language": "default",
            "children": [
                {
                    "name": "available_toilet_types",
                    "list_name": "toilet_type",
                    "label": {"english": "What type of toilets are on the premises?"},
                    "type": "select all that apply",
                    "children": [
                        {
                            "name": "pit_latrine_with_slab",
                            "label": {"english": "Pit latrine with slab"},
                        },
                        {
                            "name": "open_pit_latrine",
                            "label": {"english": "Pit latrine without slab/open pit"},
                        },
                        {
                            "name": "bucket_system",
                            "label": {"english": "Bucket system"},
                        },
                        # Removing this because select alls shouldn't need
                        # an explicit none option
                        # {
                        #    u'name': u'none',
                        #    u'label': u'None',
                        #    },
                        {"name": "other", "label": "Other"},
                    ],
                },
                {
                    "name": "available_toilet_types_other",
                    "bind": {
                        "relevant": "selected(../available_toilet_types, 'other')"
                    },
                    "label": "Specify other.",
                    "type": "text",
                },
                {
                    "name": "loop_toilet_types",
                    "type": "group",
                    "children": [
                        {
                            "name": "pit_latrine_with_slab",
                            "label": {"english": "Pit latrine with slab"},
                            "type": "group",
                            "children": [
                                {
                                    "name": "number",
                                    "label": {
                                        "english": "How many Pit latrine with slab are"
                                        " on the premises?"
                                    },
                                    "type": "integer",
                                }
                            ],
                        },
                        {
                            "name": "open_pit_latrine",
                            "label": {"english": "Pit latrine without slab/open pit"},
                            "type": "group",
                            "children": [
                                {
                                    "name": "number",
                                    "label": {
                                        "english": "How many Pit latrine without "
                                        "slab/open pit are on the premises?"
                                    },
                                    "type": "integer",
                                }
                            ],
                        },
                        {
                            "name": "bucket_system",
                            "label": {"english": "Bucket system"},
                            "type": "group",
                            "children": [
                                {
                                    "name": "number",
                                    "label": {
                                        "english": "How many Bucket system are on the"
                                        " premises?"
                                    },
                                    "type": "integer",
                                }
                            ],
                        },
                    ],
                },
                {
                    "children": [
                        {
                            "bind": {"jr:preload": "uid", "readonly": "true()"},
                            "name": "instanceID",
                            "type": "calculate",
                        }
                    ],
                    "control": {"bodyless": True},
                    "name": "meta",
                    "type": "group",
                },
            ],
        }
        self.maxDiff = None
        self.assertEqual(survey.to_json_dict(), expected_dict)

    def test_sms_columns(self):
        survey = utils.create_survey_from_fixture("sms_info", filetype=FIXTURE_FILETYPE)
        expected_dict = {
            "children": [
                {
                    "children": [
                        {
                            "label": "How old are you?",
                            "name": "age",
                            "sms_field": "q1",
                            "type": "integer",
                        },
                        {
                            "children": [
                                {"label": "no", "name": "0", "sms_option": "n"},
                                {"label": "yes", "name": "1", "sms_option": "y"},
                            ],
                            "label": "Do you have any children?",
                            "list_name": "yes_no",
                            "name": "has_children",
                            "sms_field": "q2",
                            "type": "select one",
                        },
                        {
                            "label": "What's your birth day?",
                            "name": "bday",
                            "sms_field": "q3",
                            "type": "date",
                        },
                        {
                            "label": "What is your name?",
                            "name": "name",
                            "sms_field": "q4",
                            "type": "text",
                        },
                    ],
                    "name": "section1",
                    "sms_field": "a",
                    "type": "group",
                },
                {
                    "children": [
                        {
                            "label": "May I take your picture?",
                            "name": "picture",
                            "type": "photo",
                        },
                        {
                            "label": "Record your GPS coordinates.",
                            "name": "gps",
                            "type": "geopoint",
                        },
                    ],
                    "name": "medias",
                    "sms_field": "c",
                    "type": "group",
                },
                {
                    "children": [
                        {
                            "children": [
                                {
                                    "label": "Mozilla Firefox",
                                    "name": "firefox",
                                    "sms_option": "ff",
                                },
                                {
                                    "label": "Google Chrome",
                                    "name": "chrome",
                                    "sms_option": "gc",
                                },
                                {
                                    "label": "Internet Explorer",
                                    "name": "ie",
                                    "sms_option": "ie",
                                },
                                {
                                    "label": "Safari",
                                    "name": "safari",
                                    "sms_option": "saf",
                                },
                            ],
                            "label": "What web browsers do you use?",
                            "list_name": "browsers",
                            "name": "web_browsers",
                            "sms_field": "q5",
                            "type": "select all that apply",
                        }
                    ],
                    "name": "browsers",
                    "sms_field": "b",
                    "type": "group",
                },
                {
                    "children": [
                        {
                            "label": "Phone Number",
                            "name": "phone",
                            "type": "phonenumber",
                        },
                        {"label": "Start DT", "name": "start", "type": "start"},
                        {"label": "End DT", "name": "end", "type": "end"},
                        {"label": "Send Day", "name": "today", "type": "today"},
                        {"label": "IMEI", "name": "imei", "type": "deviceid"},
                        {"label": "Hey!", "name": "nope", "type": "note"},
                    ],
                    "name": "metadata",
                    "sms_field": "meta",
                    "type": "group",
                },
                {
                    "children": [
                        {
                            "bind": {"jr:preload": "uid", "readonly": "true()"},
                            "name": "instanceID",
                            "type": "calculate",
                        }
                    ],
                    "control": {"bodyless": True},
                    "name": "meta",
                    "type": "group",
                },
            ],
            "default_language": "default",
            "id_string": "sms_info_form",
            "name": "sms_info",
            "sms_allow_media": "TRUE",
            "sms_date_format": "%Y-%m-%d",
            "sms_datetime_format": "%Y-%m-%d-%H:%M",
            "sms_keyword": "inf",
            "sms_separator": "+",
            "title": "SMS Example",
            "type": "survey",
        }
        self.assertEqual(survey.to_json_dict(), expected_dict)

    def test_style_column(self):
        survey = utils.create_survey_from_fixture(
            "style_settings", filetype=FIXTURE_FILETYPE
        )
        expected_dict = {
            "children": [
                {
                    "label": {"english": "What is your name?"},
                    "name": "your_name",
                    "type": "text",
                },
                {
                    "label": {"english": "How many years old are you?"},
                    "name": "your_age",
                    "type": "integer",
                },
                {
                    "children": [
                        {
                            "bind": {"jr:preload": "uid", "readonly": "true()"},
                            "name": "instanceID",
                            "type": "calculate",
                        }
                    ],
                    "control": {"bodyless": True},
                    "name": "meta",
                    "type": "group",
                },
            ],
            "default_language": "default",
            "id_string": "new_id",
            "name": "style_settings",
            "sms_keyword": "new_id",
            "style": "ltr",
            "title": "My Survey",
            "type": "survey",
        }
        self.assertEqual(survey.to_json_dict(), expected_dict)

    STRIP_NS_FROM_TAG_RE = re.compile(r"\{.+\}")

    def test_style_not_added_to_body_if_not_present(self):
        survey = utils.create_survey_from_fixture("settings", filetype=FIXTURE_FILETYPE)
        xml = survey.to_xml()
        # find the body tag
        root_elm = ETree.fromstring(xml.encode("utf-8"))
        body_elms = list(
            filter(
                lambda e: self.STRIP_NS_FROM_TAG_RE.sub("", e.tag) == "body",
                [c for c in root_elm.getchildren()],
            )
        )
        self.assertEqual(len(body_elms), 1)
        self.assertIsNone(body_elms[0].get("class"))

    def test_style_added_to_body_if_present(self):
        survey = utils.create_survey_from_fixture(
            "style_settings", filetype=FIXTURE_FILETYPE
        )
        xml = survey.to_xml()
        # find the body tag
        root_elm = ETree.fromstring(xml.encode("utf-8"))
        body_elms = list(
            filter(
                lambda e: self.STRIP_NS_FROM_TAG_RE.sub("", e.tag) == "body",
                [c for c in root_elm.getchildren()],
            )
        )
        self.assertEqual(len(body_elms), 1)
        self.assertEqual(body_elms[0].get("class"), "ltr")



# -*- coding: utf-8 -*-
"""
Testing simple cases for Xls2Json
"""
import codecs
import json
import os

from unittest2 import TestCase

from pyxform.tests import utils
from pyxform.xls2json import SurveyReader
from pyxform.xls2json_backends import csv_to_dict, xls_to_dict


# Nothing calls this AFAICT
def absolute_path(f, file_name):
    directory = os.path.dirname(f)
    return os.path.join(directory, file_name)


DIR = os.path.dirname(__file__)


class BasicXls2JsonApiTests(TestCase):
    maxDiff = None

    def test_simple_yes_or_no_question(self):
        filename = "yes_or_no_question.xls"
        path_to_excel_file = os.path.join(DIR, "example_xls", filename)
        # Get the xform output path:
        root_filename, ext = os.path.splitext(filename)
        output_path = os.path.join(DIR, "test_output", root_filename + ".json")
        expected_output_path = os.path.join(
            DIR, "test_expected_output", root_filename + ".json"
        )
        x = SurveyReader(path_to_excel_file, default_name="yes_or_no_question")
        x_results = x.to_json_dict()
        with codecs.open(output_path, mode="w", encoding="utf-8") as fp:
            json.dump(x_results, fp=fp, ensure_ascii=False, indent=4)
        # Compare with the expected output:
        with codecs.open(expected_output_path, "rb", encoding="utf-8") as expected_file:
            with codecs.open(output_path, "rb", encoding="utf-8") as actual_file:
                expected_json = json.load(expected_file)
                actual_json = json.load(actual_file)
                self.assertEqual(expected_json, actual_json)

    def test_hidden(self):
        x = SurveyReader(
            utils.path_to_text_fixture("hidden.xls"), default_name="hidden"
        )
        x_results = x.to_json_dict()

        expected_dict = [
            {"type": "hidden", "name": "hidden_test"},
            {
                "children": [
                    {
                        "bind": {"jr:preload": "uid", "readonly": "true()"},
                        "name": "instanceID",
                        "type": "calculate",
                    }
                ],
                "control": {"bodyless": True},
                "name": "meta",
                "type": "group",
            },
        ]
        self.assertEqual(x_results["children"], expected_dict)

    def test_gps(self):
        x = SurveyReader(utils.path_to_text_fixture("gps.xls"), default_name="gps")

        expected_dict = [
            {"type": "gps", "name": "location", "label": "GPS"},
            {
                "children": [
                    {
                        "bind": {"jr:preload": "uid", "readonly": "true()"},
                        "name": "instanceID",
                        "type": "calculate",
                    }
                ],
                "control": {"bodyless": True},
                "name": "meta",
                "type": "group",
            },
        ]

        self.assertEqual(x.to_json_dict()["children"], expected_dict)

    def test_text_and_integer(self):
        x = SurveyReader(
            utils.path_to_text_fixture("text_and_integer.xls"),
            default_name="text_and_integer",
        )

        expected_dict = [
            {
                "label": {"english": "What is your name?"},
                "type": "text",
                "name": "your_name",
            },
            {
                "label": {"english": "How many years old are you?"},
                "type": "integer",
                "name": "your_age",
            },
            {
                "children": [
                    {
                        "bind": {"jr:preload": "uid", "readonly": "true()"},
                        "name": "instanceID",
                        "type": "calculate",
                    }
                ],
                "control": {"bodyless": True},
                "name": "meta",
                "type": "group",
            },
        ]

        self.assertEqual(x.to_json_dict()["children"], expected_dict)

    def test_table(self):
        filename = "simple_loop.xls"
        path_to_excel_file = os.path.join(DIR, "example_xls", filename)
        # Get the xform output path:
        root_filename, ext = os.path.splitext(filename)
        output_path = os.path.join(DIR, "test_output", root_filename + ".json")
        expected_output_path = os.path.join(
            DIR, "test_expected_output", root_filename + ".json"
        )
        x = SurveyReader(path_to_excel_file, default_name="simple_loop")
        x_results = x.to_json_dict()
        with codecs.open(output_path, mode="w", encoding="utf-8") as fp:
            json.dump(x_results, fp=fp, ensure_ascii=False, indent=4)
        # Compare with the expected output:
        with codecs.open(expected_output_path, "rb", encoding="utf-8") as expected_file:
            with codecs.open(output_path, "rb", encoding="utf-8") as actual_file:
                expected_json = json.load(expected_file)
                actual_json = json.load(actual_file)
                self.assertEqual(expected_json, actual_json)

    def test_choice_filter_choice_fields(self):
        """
        Test that the choice filter fields appear on children field of json
        """
        choice_filter_survey = SurveyReader(
            utils.path_to_text_fixture("choice_filter_test.xlsx"),
            default_name="choice_filter_test",
        )

        expected_dict = [
            {
                "choices": [
                    {"name": "texas", "label": "Texas"},
                    {"name": "washington", "label": "Washington"},
                ],
                "type": "select one",
                "name": "state",
                "list_name": "states",
                "parameters": {},
                "label": "state",
            },
            {
                "name": "county",
                "parameters": {},
                "choice_filter": "${state}=cf",
                "label": "county",
                "itemset": "counties",
                "list_name": "counties",
                "choices": [
                    {"label": "King", "cf": "washington", "name": "king"},
                    {"label": "Pierce", "cf": "washington", "name": "pierce"},
                    {"label": "King", "cf": "texas", "name": "king"},
                    {"label": "Cameron", "cf": "texas", "name": "cameron"},
                ],
                "type": "select one",
            },
            {
                "name": "city",
                "parameters": {},
                "choice_filter": "${county}=cf",
                "label": "city",
                "itemset": "cities",
                "list_name": "cities",
                "choices": [
                    {"label": "Dumont", "cf": "king", "name": "dumont"},
                    {"label": "Finney", "cf": "king", "name": "finney"},
                    {"label": "brownsville", "cf": "cameron", "name": "brownsville"},
                    {"label": "harlingen", "cf": "cameron", "name": "harlingen"},
                    {"label": "Seattle", "cf": "king", "name": "seattle"},
                    {"label": "Redmond", "cf": "king", "name": "redmond"},
                    {"label": "Tacoma", "cf": "pierce", "name": "tacoma"},
                    {"label": "Puyallup", "cf": "pierce", "name": "puyallup"},
                ],
                "type": "select one",
            },
            {
                "control": {"bodyless": True},
                "type": "group",
                "name": "meta",
                "children": [
                    {
                        "bind": {"readonly": "true()", "jr:preload": "uid"},
                        "type": "calculate",
                        "name": "instanceID",
                    }
                ],
            },
        ]
        self.assertEqual(choice_filter_survey.to_json_dict()["children"], expected_dict)


class CsvReaderEquivalencyTest(TestCase):
    def test_equivalency(self):
        equivalent_fixtures = [
            "group",
            "loop",  # 'gps',
            "specify_other",
            "include",
            "text_and_integer",
            "include_json",
            "yes_or_no_question",
        ]
        for fixture in equivalent_fixtures:
            xls_path = utils.path_to_text_fixture("%s.xls" % fixture)
            csv_path = utils.path_to_text_fixture("%s.csv" % fixture)
            xls_inp = xls_to_dict(xls_path)
            csv_inp = csv_to_dict(csv_path)
            self.maxDiff = None
            self.assertEqual(csv_inp, xls_inp)


class UnicodeCsvTest(TestCase):
    def test_a_unicode_csv_works(self):
        """
        Simply tests that xls2json_backends.csv_to_dict does not have a problem
        with a csv with unicode characters
        """
        utf_csv_path = utils.path_to_text_fixture("utf_csv.csv")
        dict_value = csv_to_dict(utf_csv_path)
        self.assertTrue("\\ud83c" in json.dumps(dict_value))


class DefaultToSurveyTest(TestCase):
    def test_default_sheet_name_to_survey(self):
        xls_path = utils.path_to_text_fixture("survey_no_name.xlsx")
        dict_value = xls_to_dict(xls_path)
        self.assertTrue("survey" in json.dumps(dict_value))
        self.assertTrue("state" in json.dumps(dict_value))
        self.assertTrue("The State" in json.dumps(dict_value))
